from __future__ import annotations

import abc
import contextlib
import dataclasses
from configparser import RawConfigParser
from typing import TYPE_CHECKING

from loguru import logger
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from typing_extensions import Self

if TYPE_CHECKING:
    from libresvip.core.compat import Traversable

    from .base import BasePlugin, MiddlewareBase, SVSConverterBase


@dataclasses.dataclass
class BasePluginInfo(abc.ABC):
    _config: dataclasses.InitVar[RawConfigParser]
    plugin_object: BasePlugin | None = None
    name: str = dataclasses.field(init=False)
    module: str = dataclasses.field(init=False)
    version: Version = dataclasses.field(init=False)
    author: str = dataclasses.field(init=False)
    description: str = dataclasses.field(init=False)
    website: str = dataclasses.field(init=False)
    copyright: str = dataclasses.field(init=False)
    target_framework: SpecifierSet = dataclasses.field(init=False)

    def __post_init__(self, _config: RawConfigParser) -> None:
        self.module = _config.get("Core", "Module")
        self.name = _config.get("Core", "Name")
        self.version = Version(_config.get("Documentation", "Version", fallback="0.0.0"))
        self.author = _config.get("Documentation", "Author", fallback="Unknown Author")
        self.description = (
            _config.get("Documentation", "Description", fallback="")
            .encode("raw_unicode_escape")
            .decode("unicode_escape")
        )
        self.website = _config.get("Documentation", "Website", fallback="")
        self.copyright = _config.get("Documentation", "Copyright", fallback="Unknown")
        self.target_framework = SpecifierSet(
            _config.get("Documentation", "TargetFramework", fallback=">=0.0.0")
        )

    @classmethod
    def load(cls, plugfile_path: Traversable) -> Self | None:
        try:
            with plugfile_path.open(encoding="utf-8") as metafile:
                cp = RawConfigParser()
                cp.read_file(metafile)
                return cls(cp)
        except Exception:
            logger.error(f"Failed to load plugin info from {plugfile_path}")

    @classmethod
    def load_from_string(cls, content: str) -> Self | None:
        with contextlib.suppress(Exception):
            cp = RawConfigParser()
            cp.read_string(content)
            return cls(cp)

    @property
    @abc.abstractmethod
    def identifier(self) -> str: ...


@dataclasses.dataclass
class FormatProviderPluginInfo(BasePluginInfo):  # type: ignore[override]
    plugin_object: SVSConverterBase | None = None
    file_format: str = dataclasses.field(init=False)
    suffix: str = dataclasses.field(init=False)
    icon_base64: str | None = dataclasses.field(init=False)

    def __post_init__(self, _config: RawConfigParser) -> None:
        super().__post_init__(_config)
        self.file_format = _config.get("Documentation", "Format")
        self.suffix = _config.get("Documentation", "Suffix")
        self.icon_base64 = _config.get("Documentation", "IconBase64", fallback=None)

    @property
    def identifier(self) -> str:
        return self.suffix


@dataclasses.dataclass
class MiddlewarePluginInfo(BasePluginInfo):  # type: ignore[override]
    plugin_object: MiddlewareBase | None = None
    abbreviation: str = dataclasses.field(init=False)

    def __post_init__(self, _config: RawConfigParser) -> None:
        super().__post_init__(_config)
        self.abbreviation = _config.get("Documentation", "Abbreviation")

    @property
    def identifier(self) -> str:
        return self.abbreviation



from pydantic import Field

from libresvip.model.base import BaseModel
from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    SelectSingleTrackMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(EnablePitchImportationMixin, BaseModel):
    encoding: str = Field(
        default="SHIFT_JIS",
        title=_("Text encoding"),
    )


class OutputOptions(SelectSingleTrackMixin, BaseModel):
    version: float = Field(
        default=1.2,
        title=_("Version"),
        description=_("UST file version"),
    )
    encoding: str = Field(
        default="SHIFT_JIS",
        title=_("Text encoding"),
    )



from retrie.retrie import Blacklist

symbols_blacklist = Blacklist([",", ".", "?", "!", "，", "。", "？", "！"], match_substrings=True)


def get_symbol_removed_lyric(lyric: str) -> str:
    if len(lyric) > 1:
        lyric = symbols_blacklist.cleanse_text(lyric)
    return lyric



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis
from libresvip.utils.text import to_unicode

from .model import nn_grammar, nn_visitor
from .niaoniao_generator import NiaoniaoGenerator
from .niaoniao_parser import NiaoNiaoParser
from .options import InputOptions, OutputOptions
from .template import render_nn


class NiaoNiaoConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        tree = nn_grammar.parse(to_unicode(path.read_bytes()))
        nn_project = nn_visitor.visit(tree)
        return NiaoNiaoParser(options).parse_project(nn_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        if len(project.song_tempo_list) != 1:
            project = reset_time_axis(project, options.tempo)
        nn_project = NiaoniaoGenerator(options).generate_project(project)
        render_nn(nn_project, path)



from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class InputOptions(BaseModel):
    wave_to_singing: bool = Field(True, title=_("Convert wave pattern to singing pattern"))
    use_edited_pitch: bool = Field(True, title=_("Use edited pitch curve"))
    use_edited_dynamics: bool = Field(True, title=_("Use edited dynamics curve"))
    import_dynamics: bool = Field(False, title=_("Import dynamics curve"))
    import_formant: bool = Field(False, title=_("Import formant curve"))
    import_breath: bool = Field(False, title=_("Import breath curve"))



from typing import Final

DEFAULT_PHONEME: Final[str] = "r,a"
TIME_UNIT_AS_TICKS_PER_BPM: Final[float] = 4.8 / 120
OCTAVE_OFFSET: Final[int] = -1
TICK_RATE: Final[float] = 2.0
MIN_DATA_LENGTH: Final[int] = 500
TEMP_VALUE_AS_NULL: Final[float] = -1.0



import collections
import dataclasses
import math
import operator

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import Note, Project, SingingTrack, SongTempo, TimeSignature
from libresvip.model.point import Point
from libresvip.utils.translation import gettext_lazy as _

from .model import VxFile, VxPitchData, VxTrack
from .options import InputOptions


@dataclasses.dataclass
class VxBetaParser:
    options: InputOptions
    ticks_per_beat: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    @property
    def tick_rate(self) -> float:
        if self.ticks_per_beat is not None:
            return TICKS_IN_BEAT / self.ticks_per_beat
        return 1

    def parse_project(self, vx_project: VxFile) -> Project:
        self.ticks_per_beat = vx_project.ticks_per_beat
        self._convert_delta_to_cumulative(vx_project.tracks)
        time_signature_list = []
        if len(vx_project.tracks) > 0:
            master_track = vx_project.tracks[0]
            time_signature_list.extend(self.parse_time_signatures(master_track))
        song_tempo_list = self.parse_tempos(vx_project.tracks)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=time_signature_list,
            track_list=self.parse_tracks(vx_project.tracks),
        )

    @staticmethod
    def _convert_delta_to_cumulative(tracks: list[VxTrack]) -> None:
        for track in tracks:
            tick = 0
            for event in track.events:
                event.time += tick
                tick = event.time

    def parse_tempos(self, tracks: list[VxTrack]) -> list[SongTempo]:
        tempos: list[SongTempo] = []

        # traversing
        for track in tracks:
            for event in track.events:
                if event.type == "set_tempo":
                    # convert tempo to BPM
                    tempo = round(event.tempo, 3)
                    tick = round(event.time * self.tick_rate)
                    last_tempo = tempos[-1].bpm if tempos else None
                    if tempo != last_tempo:
                        tempos.append(SongTempo(position=tick, bpm=tempo))
        if not tempos:
            # default bpm
            show_warning(_("No tempo labels found in the imported project."))
            tempos.append(SongTempo(position=0, bpm=self.options.default_bpm))
        else:
            tempos.sort(key=operator.attrgetter("position"))
        return tempos

    def parse_time_signatures(self, master_track: VxTrack) -> list[TimeSignature]:
        # no default
        time_signature_changes: list[TimeSignature] = []

        # traversing
        prev_ticks = 0
        measure = 0
        for event in master_track.events:
            if event.type == "time_signature":
                tick = event.time
                if not time_signature_changes:
                    tick_in_full_note = 4 * self.ticks_per_beat
                else:
                    tick_in_full_note = round(
                        time_signature_changes[-1].bar_length(self.ticks_per_beat)
                    )
                measure += (tick - prev_ticks) / tick_in_full_note
                ts_obj = TimeSignature(
                    bar_index=math.floor(measure),
                    numerator=event.numerator,
                    denominator=event.denominator,
                )
                time_signature_changes.append(ts_obj)
                prev_ticks = tick
        if not time_signature_changes or time_signature_changes[0].bar_index > 0:
            time_signature_changes.insert(0, TimeSignature(bar_index=0, numerator=4, denominator=4))
        self.first_bar_length = round(time_signature_changes[0].bar_length())
        return time_signature_changes

    def parse_track(self, track: VxTrack) -> SingingTrack:
        lyrics: dict[int, str] = collections.defaultdict(lambda: "l-aa")
        prev_index = None
        for event in track.events:
            if event.type == "lyrics":
                seq_stat = (event.seq_stat - 16) // 64
                if seq_stat == 0:
                    lyrics[event.seq_num] = event.text
                    prev_index = None
                elif seq_stat == 1:
                    lyrics[event.seq_num] = event.text
                    prev_index = event.seq_num
                elif seq_stat == 2 and prev_index is not None:
                    lyrics[prev_index] += event.text
                elif seq_stat == 3 and prev_index is not None:
                    lyrics[prev_index] += event.text
                    prev_index = None
        last_note_on = None
        notes: list[Note] = []
        for event in track.events:
            if event.type == "note_on":
                last_note_on = event
            elif event.type == "note_off" and last_note_on:
                notes.append(
                    Note(
                        start_pos=round(last_note_on.time * self.tick_rate),
                        length=round((event.time - last_note_on.time) * self.tick_rate),
                        lyric=lyrics[len(notes)],
                        key_number=event.note,
                    )
                )
                last_note_on = None
        singing_track = SingingTrack(
            title="".join(event.name for event in track.title_parts),
            note_list=notes,
        )
        if self.options.import_pitch and (
            buffer := "".join(event.text for event in track.events if event.type == "metadata")
        ):
            pitch_data = VxPitchData.model_validate_json(buffer)
            pitch_points = [Point.start_point()]
            prev_pos = None
            for point in pitch_data.time_based_pitch_sequence.pitch_sequence:
                if prev_pos and point.position - prev_pos > 1 and pitch_points:
                    pitch_points.append(
                        Point(
                            x=pitch_points[-1].x,
                            y=-100,
                        )
                    )
                if pitch_points[-1].y == -100:
                    pitch_points.append(
                        Point(
                            x=int(
                                self.synchronizer.get_actual_ticks_from_secs(
                                    point.position
                                    * pitch_data.time_based_pitch_sequence.time_frame_period_seconds
                                )
                            )
                            + self.first_bar_length,
                            y=-100,
                        )
                    )
                pitch_points.append(
                    Point(
                        x=int(
                            self.synchronizer.get_actual_ticks_from_secs(
                                point.position
                                * pitch_data.time_based_pitch_sequence.time_frame_period_seconds
                            )
                        )
                        + self.first_bar_length,
                        y=int(point.pitch) + 6900,
                    )
                )
                prev_pos = point.position
            if len(pitch_points) > 1:
                pitch_points.append(
                    Point(
                        x=pitch_points[-1].x,
                        y=-100,
                    )
                )
            pitch_points.append(Point.end_point())
            singing_track.edited_params.pitch.points.root = pitch_points
        return singing_track

    def parse_tracks(self, vx_tracks: list[VxTrack]) -> list[SingingTrack]:
        return [self.parse_track(track) for track in vx_tracks]



import atexit
import io
from contextlib import ExitStack

from loguru import _logger
from typing_extensions import Self

warning_logger = _logger.Logger(
    core=_logger.Core(),
    exception=None,
    depth=0,
    record=False,
    lazy=False,
    colors=False,
    raw=False,
    capture=True,
    patchers=[],
    extra={},
)


def show_warning(message: str) -> None:
    warning_logger.warning(message)


class CatchWarnings:
    def __init__(self) -> None:
        self._handler_id = None
        self._exit_stack: ExitStack | None = None
        self._output = io.StringIO()
        self.output = ""

    def __enter__(self) -> Self:
        self._handler_id = warning_logger.add(
            self._output,
            format="{extra[handler_id]}|{message}",
            level="WARNING",
        )
        with ExitStack() as exit_stack:
            exit_stack.enter_context(warning_logger.contextualize(handler_id=self._handler_id))
            self._exit_stack = exit_stack.pop_all()
        return self

    def __exit__(self, *args: object) -> None:
        msg_prefix = f"{self._handler_id}|"
        if self._handler_id is not None:
            self._exit_stack.close()
            warning_logger.remove(self._handler_id)
            self._handler_id = None
        self._output.seek(0)
        lines = self._output.readlines()
        self.output = "\n".join(
            line.removeprefix(msg_prefix).rstrip() for line in lines if line.startswith(msg_prefix)
        )


atexit.register(warning_logger.remove)



import gettext
from typing import Any

from PySide6.QtCore import QLocale, QObject, QTranslator, Signal, Slot
from PySide6.QtQml import QmlElement

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from libresvip.core.config import Language, config_path, settings
from libresvip.extension.manager import get_translation
from libresvip.utils import translation

from .application import app, qml_engine

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


class GettextTranslator(QTranslator):
    def load_translation(self, lang: str) -> None:
        try:
            translation.singleton_translation = get_translation(lang=lang)
        except OSError:
            translation.singleton_translation = gettext.NullTranslations()

    def translate(
        self,
        context: str,
        source_text: str,
        disambiguation: bytes | None = None,
        n: int = 0,
    ) -> str:
        if translation.singleton_translation is not None and source_text.strip():
            if (
                contextual_text := translation.singleton_translation.pgettext(context, source_text)
            ) != source_text:
                return contextual_text
            return translation.singleton_translation.gettext(source_text)
        return source_text


class SingletonMetaObject(type(QObject)):  # type: ignore[misc]
    _instance = None

    def __new__(cls, name: str, bases: tuple[type], attrs: dict[str, Any]) -> type[QObject]:
        if not cls._instance:
            cls._instance = super().__new__(cls, name, bases, attrs)
        return cls._instance


@QmlElement
class LocaleSwitcher(QObject, metaclass=SingletonMetaObject):
    translator_initialized = Signal()

    def __init__(self) -> None:
        super().__init__()
        self._translator_initialized = False
        self.translator = GettextTranslator()
        if not config_path.exists():
            sys_locale = QLocale.system().name()
            settings.language = Language.from_locale(sys_locale)

    @Slot(result=str)
    def get_language(self) -> str:
        return settings.language.value

    @Slot(str)
    def switch_language(self, lang: str) -> None:
        if lang:
            self.translator.load_translation(lang)
            app.install_translator(self.translator)
            qml_engine.retranslate()
            settings.language = Language.from_locale(lang)
        else:
            app.remove_translator(self.translator)
            settings.language = Language.ENGLISH
        if not self._translator_initialized:
            self._translator_initialized = True
            self.translator_initialized.emit()



_colors = [
    "#b17ef3",
    "#da9de3",
    "#dd88d5",
    "#f57db7",
    "#dd8585",
    "#ebac99",
    "#f4a070",
    "#f1c970",
    "#c0cf63",
    "#99f07b",
    "#70d4b1",
    "#91bcdc",  # start
    "#96afef",
    "#979ae4",
]


def count_color() -> int:
    return len(_colors)


def get_color(index: int) -> str:
    return _colors[index % len(_colors)]



# © 2018 Gerard Marull-Paretas <gerard@teslabs.com>
# © 2014 Mark Harviston <mark.harviston@gmail.com>
# © 2014 Arve Knudsen <arve.knudsen@gmail.com>
# BSD License

# mypy: disable-error-code="attr-defined"

"""Windows specific Quamash functionality."""

from __future__ import annotations

import _overlapped
import _winapi
import asyncio
import math
import sys
from asyncio import windows_events
from typing import IO, TYPE_CHECKING, Any

from loguru import logger
from PySide6 import QtCore

if TYPE_CHECKING:
    import socket
    from _winapi import Overlapped
    from asyncio import windows_utils
    from collections.abc import Callable

    from _typeshed import Incomplete, ReadableBuffer, WriteableBuffer

    ProactorEvent = tuple[
        asyncio.Future[Any],
        Callable[[Incomplete, Incomplete, Overlapped], Any],
        Incomplete,
        Incomplete,
        Overlapped,
    ]

from ._common import make_signaller

UINT32_MAX = 0xFFFFFFFF


class _ProactorEventLoop(asyncio.ProactorEventLoop):  # type: ignore[name-defined]
    """Proactor based event loop."""

    def __init__(self) -> None:
        super().__init__(_IocpProactor())

        self.__event_signaller = make_signaller(list)
        self.__event_signal = self.__event_signaller.signal
        self.__event_signal.connect(self._process_events)
        self.__event_poller = _EventPoller(self.__event_signal)

    def _process_events(self, events: list[ProactorEvent]) -> None:
        """Process events from proactor."""
        for event in events:
            self._process_event(event)

    def _process_event(self, event: ProactorEvent) -> None:
        f, callback, transferred, key, ov = event
        try:
            logger.debug("Invoking event callback {}", callback)
            value = callback(transferred, key, ov)
        except OSError as e:
            logger.exception("Event callback failed", exc_info=sys.exc_info())
            if not f.done():
                f.set_exception(e)
        else:
            if not f.cancelled():
                f.set_result(value)

    def _before_run_forever(self) -> None:
        self.__event_poller.start(self._proactor)

    def _after_run_forever(self) -> None:
        self.__event_poller.stop()


class _IocpProactor(windows_events.IocpProactor):  # type: ignore[name-defined]
    def __init__(self) -> None:
        self.__events: list[ProactorEvent] = []
        super().__init__()
        self._lock = QtCore.QMutex()

    def select(self, timeout: float | None = None) -> list[ProactorEvent]:  # type: ignore[override]
        """Override in order to handle events in a threadsafe manner."""
        if not self.__events:
            self._poll(timeout)
        tmp = self.__events
        self.__events = []
        return tmp

    def close(self) -> None:
        logger.debug("Closing")
        super().close()

    # Wrap all I/O submission methods to acquire the internal lock first; listed
    # in the order they appear in the base class source code.

    def recv(self, conn: socket.socket, nbytes: int, flags: int = 0) -> asyncio.Future[bytes]:
        with QtCore.QMutexLocker(self._lock):
            return super().recv(conn, nbytes, flags)

    def recv_into(
        self, conn: socket.socket, buf: WriteableBuffer, flags: int = 0
    ) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().recv_into(conn, buf, flags)

    def recvfrom(
        self, conn: socket.socket, nbytes: int, flags: int = 0
    ) -> asyncio.Future[tuple[bytes, socket._RetAddress]]:
        with QtCore.QMutexLocker(self._lock):
            return super().recvfrom(conn, nbytes, flags)

    if sys.version_info >= (3, 11):

        def recvfrom_into(
            self, conn: socket.socket, buf: WriteableBuffer, flags: int = 0
        ) -> asyncio.Future[tuple[int, socket._RetAddress]]:
            with QtCore.QMutexLocker(self._lock):
                return super().recvfrom_into(conn, buf, flags)

    def sendto(
        self,
        conn: socket.socket,
        buf: ReadableBuffer,
        flags: int = 0,
        addr: socket._RetAddress | None = None,
    ) -> asyncio.Future[int]:
        with QtCore.QMutexLocker(self._lock):
            return super().sendto(conn, buf, flags, addr)

    def send(
        self, conn: socket.socket, buf: WriteableBuffer, flags: int = 0
    ) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().send(conn, buf, flags)

    def accept(self, listener: socket.socket) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().accept(listener)

    def connect(
        self,
        conn: socket.socket,
        address: tuple[Incomplete, Incomplete]
        | tuple[Incomplete, Incomplete, Incomplete, Incomplete],
    ) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().connect(conn, address)

    def sendfile(
        self, sock: socket.socket, file: IO[bytes], offset: int, count: int
    ) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().sendfile(sock, file, offset, count)

    def accept_pipe(self, pipe: socket.socket) -> asyncio.Future[Any]:
        with QtCore.QMutexLocker(self._lock):
            return super().accept_pipe(pipe)

    # connect_pipe() does not actually use the delayed completion machinery.

    def _wait_for_handle(
        self,
        handle: windows_utils.PipeHandle,  # type: ignore[name-defined]
        timeout: int,
        _is_cancel: bool,
    ) -> bool:
        with QtCore.QMutexLocker(self._lock):
            return super()._wait_for_handle(  # type: ignore[misc]
                handle, timeout, _is_cancel
            )

    def _poll(self, timeout: float | None = None) -> None:
        """Override in order to handle events in a threadsafe manner."""
        if timeout is None:
            ms = UINT32_MAX  # wait for eternity
        elif timeout < 0:
            msg = "negative timeout"
            raise ValueError(msg)
        else:
            # GetQueuedCompletionStatus() has a resolution of 1 millisecond,
            # round away from zero to wait *at least* timeout seconds.
            ms = math.ceil(timeout * 1e3)
            if ms >= UINT32_MAX:
                msg = "timeout too big"
                raise ValueError(msg)

        while True:
            status = _overlapped.GetQueuedCompletionStatus(self._iocp, ms)
            if status is None:
                break
            ms = 0

            with QtCore.QMutexLocker(self._lock):
                err, transferred, key, address = status
                try:
                    f, ov, obj, callback = self._cache.pop(address)
                except KeyError:
                    # key is either zero, or it is used to return a pipe
                    # handle which should be closed to avoid a leak.
                    if key not in (0, _overlapped.INVALID_HANDLE_VALUE):
                        _winapi.CloseHandle(key)
                    continue

                if obj in self._stopped_serving:
                    f.cancel()
                # Futures might already be resolved or cancelled
                elif not f.done():
                    self.__events.append((f, callback, transferred, key, ov))

        # Remove unregistered futures
        for ov in self._unregistered:
            self._cache.pop(ov.address, None)
        self._unregistered.clear()


class _EventPoller:
    sig_events: QtCore.SignalInstance

    """Polling of events in separate thread."""

    def __init__(self, sig_events: QtCore.SignalInstance) -> None:
        self.sig_events = sig_events

    def start(self, proactor: _IocpProactor) -> None:
        logger.debug("Starting (proactor: {})...", proactor)
        self.__worker = _EventWorker(proactor, self)
        self.__worker.start()

    def stop(self) -> None:
        logger.debug("Stopping worker thread...")
        self.__worker.stop()


class _EventWorker(QtCore.QThread):
    def __init__(self, proactor: _IocpProactor, parent: _EventPoller) -> None:
        super().__init__()

        self.__stop = False
        self.__proactor = proactor
        self.__sig_events: QtCore.SignalInstance = parent.sig_events
        self.__semaphore = QtCore.QSemaphore()

    def start(
        self, /, priority: QtCore.QThread.Priority = QtCore.QThread.Priority.NormalPriority
    ) -> None:
        super().start(priority=priority)
        self.__semaphore.acquire()

    def stop(self) -> None:
        self.__stop = True
        # Wait for thread to end
        self.wait()

    def run(self) -> None:
        logger.debug("Thread started")
        self.__semaphore.release()

        while not self.__stop:
            if events := self.__proactor.select(0.01):  # type: ignore[arg-type]
                logger.debug("Got events from poll: {}", events)
                self.__sig_events.emit(events)

        logger.debug("Exiting thread")



import math
import struct
from typing import BinaryIO, Union

import more_itertools
from construct import (
    Byte,
    BytesInteger,
    Computed,
    Construct,
    Container,
    CString,
    Float64l,
    GreedyBytes,
    Int64sl,
    LazyBound,
    Prefixed,
    PrefixedArray,
    SizeofError,
    Struct,
    Switch,
    this,
)
from construct import Enum as CSEnum
from construct import Path as CSPath
from construct_typed import Context
from typing_extensions import Never

Int32sl = BytesInteger(4, swapped=True, signed=True)
Variant = bool | int | float | str | bytes | list["Variant"]
Node = dict[str, Union[Variant, "Node", list["Node"]]]

JUCEVarTypes = CSEnum(
    Byte,
    INT=1,
    BOOL_TRUE=2,
    BOOL_FALSE=3,
    DOUBLE=4,
    STRING=5,
    INT64=6,
    ARRAY=7,
    BINARY=8,
    UNDEFINED=9,
)


class JUCECompressedIntStruct(Construct):
    def _sizeof(self, context: Context, path: CSPath) -> Never:
        msg = "JUCECompressedInt has no static size"
        raise SizeofError(msg)

    def _parse(self, stream: BinaryIO, context: Context, path: CSPath) -> int:
        byte = stream.read(1)
        if not byte:
            raise EOFError
        width = struct.unpack("<B", byte)[0]
        return int.from_bytes(stream.read(width), "little", signed=False)

    def _build(self, obj: int, stream: BinaryIO, context: Context, path: CSPath) -> int:
        if obj < 0:
            msg = "Negative numbers not supported"
            raise ValueError(msg)
        width = max(math.ceil(obj.bit_length() / 8), 1)
        try:
            content = obj.to_bytes(width, "little", signed=False)
            stream.write(struct.pack("<B", width))
            stream.write(content)
        except OverflowError as e:
            msg = "Number too large to be compressed"
            raise ValueError(msg) from e
        else:
            return obj


JUCECompressedInt = JUCECompressedIntStruct()


JUCEVariant: Container = Prefixed(
    JUCECompressedInt,
    Struct(
        "type" / JUCEVarTypes,
        "value"
        / Switch(
            this.type,
            {
                "INT": Int32sl,
                "BOOL_TRUE": Computed(lambda ctx: True),
                "BOOL_FALSE": Computed(lambda ctx: False),
                "DOUBLE": Float64l,
                "STRING": CString("utf-8"),
                "INT64": Int64sl,
                "ARRAY": PrefixedArray(
                    JUCECompressedInt,
                    LazyBound(lambda: JUCEVariant),
                ),
                "BINARY": GreedyBytes,
            },
        ),
    ),
)

JUCENamedVariant: Container = Struct(
    "name" / CString("utf-8"),
    "data" / JUCEVariant,
)

JUCENode: Container = Struct(
    "name" / CString("utf-8"),
    "attrs" / PrefixedArray(JUCECompressedInt, JUCENamedVariant),
    "children" / PrefixedArray(JUCECompressedInt, LazyBound(lambda: JUCENode)),
)


def build_variant(variant: Container) -> Variant:
    if isinstance(variant.value, list):
        return [build_variant(x) for x in variant.value]
    return variant.value


def build_tree_dict(node: Container) -> Node:
    attr_dict: Node = {
        attr.name: (
            build_tree_dict(JUCENode.parse(attr.data.value))
            if isinstance(attr.data.value, bytes)
            else build_variant(attr.data)
        )
        for attr in node.attrs
    }
    buckets = more_itertools.bucket(
        (build_tree_dict(child) for child in node.children),
        key=lambda item: next(iter(item.keys())),
    )
    children_dict: Node = {
        key: [next(iter(item.values())) for item in buckets[key]] for key in buckets
    }
    return {node.name: children_dict | attr_dict}



import dataclasses

from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)

from .model import (
    VogenNote,
    VogenProject,
    VogenTrack,
)
from .options import InputOptions


@dataclasses.dataclass
class VogenParser:
    options: InputOptions

    def parse_project(self, vogen_project: VogenProject) -> Project:
        return Project(
            song_tempo_list=self.parse_tempos(vogen_project.bpm0),
            time_signature_list=self.parse_time_signatures(vogen_project.time_sig0),
            track_list=self.parse_tracks(vogen_project.utts),
        )

    def parse_tempos(self, bpm0: float) -> list[SongTempo]:
        return [SongTempo(position=0, bpm=bpm0)]

    def parse_time_signatures(self, time_sig0: str) -> list[TimeSignature]:
        numerator, _, denominator = time_sig0.partition("/")
        return [TimeSignature(numerator=int(numerator), denominator=int(denominator))]

    def parse_tracks(self, utts: list[VogenTrack]) -> list[SingingTrack]:
        return [
            SingingTrack(
                title=utt.name,
                ai_singer_name=utt.singer_id,
                note_list=self.parse_notes(utt.notes),
            )
            for utt in utts
        ]

    def parse_notes(self, notes: list[VogenNote]) -> list[Note]:
        return [
            Note(
                start_pos=note.on,
                length=note.dur,
                lyric=note.lyric,
                pronunciation=note.rom,
                key_number=note.pitch,
            )
            for note in notes
        ]



import dataclasses
import re
from importlib.resources import files

from bidict import bidict

from libresvip.core.compat import json

from .msnrbf.xstudio_models import (
    XSNoteHeadTagEnum,
    XSReverbPresetEnum,
)


@dataclasses.dataclass
class OpenSvipSingers:
    singers: bidict[str, str] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        singers_data_path = files("libresvip.plugins.svip") / "singers.json"
        self.singers = bidict(json.loads(singers_data_path.read_text(encoding="utf-8")))

    def get_name(self, id_: str) -> str:
        if id_ in self.singers:
            return self.singers[id_]
        return f"$({id_})" if re.match(r"[FM]\d+", id_) is not None else ""

    def get_id(self, name: str) -> str:
        if name in self.singers.inverse:
            return self.singers.inverse[name]
        return name[2:-1] if re.match(r"\$\([FM]\d+\)", name) is not None else ""


opensvip_singers = OpenSvipSingers()

svip_reverb_presets = bidict(
    {
        "干声": XSReverbPresetEnum.NONE,
        "浮光": XSReverbPresetEnum.DEFAULT,
        "午后": XSReverbPresetEnum.SMALLHALL1,
        "月光": XSReverbPresetEnum.MEDIUMHALL1,
        "水晶": XSReverbPresetEnum.LARGEHALL1,
        "汽水": XSReverbPresetEnum.SMALLROOM1,
        "夜莺": XSReverbPresetEnum.MEDIUMROOM1,
        "大梦": XSReverbPresetEnum.LONGREVERB2,
    }
)


svip_note_head_tags = bidict(
    {
        None: XSNoteHeadTagEnum.NoTag,
        "0": XSNoteHeadTagEnum.SilTag,
        "V": XSNoteHeadTagEnum.SpTag,
    }
)



from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.lyric_phoneme.japanese import (
    to_hiragana,
    to_katakana,
    to_romaji,
)
from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import legato_chars
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project, SingingTrack

from .options import ProcessOptions, PronounciationConversionOptions


class PronounciationMiddleware(plugin_base.MiddlewareBase):
    def process(self, project: Project, options: ProcessOptions) -> Project:
        if options.mode != PronounciationConversionOptions.NONE:
            if options.mode == PronounciationConversionOptions.TO_ROMAJI:
                lyric_transformer = to_romaji
            elif options.mode == PronounciationConversionOptions.TO_KATAKANA:
                lyric_transformer = to_katakana
            elif options.mode == PronounciationConversionOptions.TO_HIRAGANA:
                lyric_transformer = to_hiragana
            else:

                def lyric_transformer(lyric: str) -> str:  # type: ignore[misc]
                    return " ".join(get_pinyin_series([lyric]))

            for track in project.track_list:
                if isinstance(track, SingingTrack):
                    for note in track.note_list:
                        note.lyric = (
                            "-" if note.lyric in legato_chars else lyric_transformer(note.lyric)
                        )
        return project



import dataclasses

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve

from .constants import (
    MODE1_PITCH_SAMPLING_INTERVAL_TICK,
)
from .resampling import dot_resampled


@dataclasses.dataclass
class UtauMode1NotePitchData:
    pitch_points: list[int] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class UtauMode1TrackPitchData:
    notes: list[UtauMode1NotePitchData | None] = dataclasses.field(default_factory=list)


def pitch_from_utau_mode1_track(
    pitch_data: UtauMode1TrackPitchData,
    synchronizer: TimeSynchronizer,
    notes: list[Note],
) -> ParamCurve:
    pitch_points: list[Point] = []
    for note, note_pitch in zip(notes, pitch_data.notes):
        if note_pitch is not None and len(note_pitch.pitch_points):
            pitch_points.extend(
                Point(
                    x=note.start_pos + index * MODE1_PITCH_SAMPLING_INTERVAL_TICK,
                    y=value,
                )
                for index, value in enumerate(note_pitch.pitch_points)
            )
    pitch_simulator = PitchSimulator(
        synchronizer=synchronizer,
        portamento=PortamentoPitch.no_portamento(),
        note_list=notes,
    )
    return RelativePitchCurve(
        lower_bound=notes[0].start_pos,
        upper_bound=notes[-1].end_pos,
    ).to_absolute(pitch_points, pitch_simulator)


def pitch_to_utau_mode1_track(pitch: ParamCurve, notes: list[Note]) -> UtauMode1TrackPitchData:
    note_pitch_data: list[UtauMode1NotePitchData | None] = []
    for note in notes:
        data = [
            point
            for point in pitch.points.root
            if note.start_pos <= point.x - 1920 < note.end_pos and point.y != -100
        ]
        if not data:
            note_pitch_data.append(UtauMode1NotePitchData())
            continue
        resampled_data = dot_resampled(data, MODE1_PITCH_SAMPLING_INTERVAL_TICK)
        pitch_points = [round(pitch) - note.key_number * 100 for _, pitch in resampled_data]
        note_pitch_data.append(UtauMode1NotePitchData(pitch_points))
    return UtauMode1TrackPitchData(note_pitch_data)



import dataclasses
import functools
import math
from collections.abc import Iterable

import portion
from more_itertools import convolve

from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.utils.music_math import linear_interpolation

from .model import AcepNote, AcepVibrato


@dataclasses.dataclass
class NoteInSeconds:
    semitone: int
    start: float = 0.0
    end: float = 0.0


def _convolve(note_list: list[NoteInSeconds]) -> list[float]:
    total_points = round(1000 * (note_list[-1].end + 0.12)) + 1
    init_values = [0.0] * total_points
    note_index = 0
    for i in range(total_points):
        init_values[i] = note_list[note_index].semitone
        if note_index < len(note_list) - 1:
            ts = 0.001 * i
            if ts >= 0.5 * (note_list[note_index].end + note_list[note_index + 1].start):
                note_index += 1
    kernel = [0.0] * 119
    for i in range(119):
        ts = 0.001 * (i - 59)
        kernel[i] = math.cos(math.pi * ts / 0.12)
    kernel_sum = sum(kernel)
    for i in range(119):
        kernel[i] /= kernel_sum
    return list(convolve(init_values, kernel))[59:-59]


def acep_vibrato_value_curve(seconds: float, vibrato_start: float, vibrato: AcepVibrato) -> float:
    return (
        math.sin(math.pi * (2 * (seconds - vibrato_start) * vibrato.frequency - vibrato.phase))
        * vibrato.amplitude
        * 0.5
    )


@dataclasses.dataclass
class BasePitchCurve:
    notes: dataclasses.InitVar[Iterable[AcepNote]]
    synchronizer: dataclasses.InitVar[TimeSynchronizer]
    tick_offset: dataclasses.InitVar[int] = 0
    vibrato_value_interval_dict: PiecewiseIntervalDict = dataclasses.field(
        default_factory=PiecewiseIntervalDict
    )
    vibrato_coef_interval_dict: PiecewiseIntervalDict = dataclasses.field(
        default_factory=PiecewiseIntervalDict
    )
    values_in_semitone: list[float] = dataclasses.field(default_factory=list)

    def __post_init__(
        self,
        notes: Iterable[AcepNote],
        synchronizer: TimeSynchronizer,
        tick_offset: int,
    ) -> None:
        note_list = []
        for note in notes:
            note_end = synchronizer.get_actual_secs_from_ticks(note.pos + note.dur + tick_offset)
            note_list.append(
                NoteInSeconds(
                    start=synchronizer.get_actual_secs_from_ticks(note.pos + tick_offset),
                    end=note_end,
                    semitone=note.pitch,
                )
            )
            if note.vibrato is not None:
                vibrato_start = synchronizer.get_actual_secs_from_ticks(
                    int(note.pos + note.vibrato.start_pos + tick_offset)
                )
                vibrato_duration = note_end - vibrato_start
                self.vibrato_value_interval_dict[portion.closed(vibrato_start, note_end)] = (
                    functools.partial(
                        acep_vibrato_value_curve,
                        vibrato_start=vibrato_start,
                        vibrato=note.vibrato,
                    )
                )
                attack_time = vibrato_start + note.vibrato.attack_ratio * vibrato_duration
                release_time = note_end - note.vibrato.release_ratio * vibrato_duration
                if note.vibrato.release_ratio:
                    self.vibrato_coef_interval_dict[portion.openclosed(release_time, note_end)] = (
                        functools.partial(
                            linear_interpolation,  # type: ignore[call-arg]
                            start=(release_time, note.vibrato.release_level),
                            end=(note_end, 0),
                        )
                    )
                self.vibrato_coef_interval_dict[portion.closed(attack_time, release_time)] = (
                    functools.partial(
                        linear_interpolation,  # type: ignore[call-arg]
                        start=(attack_time, note.vibrato.attack_level),
                        end=(release_time, note.vibrato.release_level),
                    )
                )
                if note.vibrato.attack_ratio:
                    self.vibrato_coef_interval_dict[
                        portion.closedopen(vibrato_start, attack_time)
                    ] = functools.partial(
                        linear_interpolation,  # type: ignore[call-arg]
                        start=(vibrato_start, 0),
                        end=(attack_time, note.vibrato.attack_level),
                    )
        self.values_in_semitone = _convolve(note_list)

    def semitone_value_at(self, seconds: float) -> float:
        position = 1000 * max(0.0, seconds)
        left_index = math.floor(position)
        lambda_ = position - left_index
        clipped_left_index = min(int(left_index), len(self.values_in_semitone) - 1)
        clipped_right_index = min(clipped_left_index + 1, len(self.values_in_semitone) - 1)
        pitch_value = (1 - lambda_) * self.values_in_semitone[
            clipped_left_index
        ] + lambda_ * self.values_in_semitone[clipped_right_index]
        if (vibrato_value := self.vibrato_value_interval_dict.get(seconds)) is not None:
            vibrato_value *= self.vibrato_coef_interval_dict.get(seconds, 0)
            pitch_value += vibrato_value
        return pitch_value



import dataclasses
import functools
import math
import pathlib
import secrets

import more_itertools
import portion

from libresvip.core.constants import DEFAULT_BPM, DEFAULT_PHONEME, TICKS_IN_BEAT
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Project,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.music_math import linear_interpolation

from .model import (
    VOXFactoryAudioClip,
    VOXFactoryAudioData,
    VOXFactoryAudioTrack,
    VOXFactoryNote,
    VOXFactoryProject,
    VOXFactoryTrack,
    VOXFactoryVocalClip,
    VOXFactoryVocalTrack,
)
from .options import OutputOptions


@dataclasses.dataclass
class VOXFactoryGenerator:
    options: OutputOptions
    prefix: str = dataclasses.field(init=False)
    audio_paths: dict[str, pathlib.Path] = dataclasses.field(default_factory=dict)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> VOXFactoryProject:
        self.prefix = secrets.token_hex(5)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        track_bank, audio_data_bank = self.generate_tracks(project.track_list)
        return VOXFactoryProject(
            tempo=self.generate_tempo(project.song_tempo_list),
            time_signature=self.generate_time_signature(project.time_signature_list),
            track_bank=track_bank,
            track_order=sorted(track_bank.keys()),
            audio_data_bank=audio_data_bank,
        )

    def generate_tempo(self, tempos: list[SongTempo]) -> float:
        return tempos[0].bpm if tempos else DEFAULT_BPM

    def generate_time_signature(self, time_signatures: list[TimeSignature]) -> list[int]:
        if time_signatures:
            return [time_signatures[0].numerator, time_signatures[0].denominator]
        else:
            return [4, 4]

    def generate_tracks(
        self, tracks: list[Track]
    ) -> tuple[dict[str, VOXFactoryTrack], dict[str, VOXFactoryAudioData]]:
        track_bank = {}
        audio_data_bank = {}
        for i, track in enumerate(tracks):
            if isinstance(track, InstrumentalTrack):
                audio_path = pathlib.Path(track.audio_file_path)
                if (track_info := audio_track_info(track.audio_file_path)) is not None:
                    source_audio_data_key = f"{self.prefix}-au{i}{audio_path.suffix}"
                    self.audio_paths[source_audio_data_key] = audio_path
                    audio_data_bank[source_audio_data_key] = VOXFactoryAudioData(
                        sample_rate=44100,
                        sample_length=int(track_info.duration * 44100 / 1000),
                        number_of_channels=track_info.channel_s,
                    )
                    clip_bank = {
                        f"{self.prefix}-cl0": VOXFactoryAudioClip(
                            name=audio_path.stem,
                            offset_quarter=0,
                            start_quarter=track.offset / TICKS_IN_BEAT,
                            length=track_info.duration / 1000,
                            source_audio_data_key=source_audio_data_key,
                        )
                    }
                    clip_order = [f"{self.prefix}-cl0"]
                    track_bank[f"{self.prefix}-tr{i}"] = VOXFactoryAudioTrack(
                        clip_bank=clip_bank,
                        clip_order=clip_order,
                        name=track.title,
                        mute=track.mute,
                        solo=track.solo,
                        pan=track.pan,
                    )
            else:
                clip_bank = self.generate_notes(track.note_list, track.edited_params)
                clip_order = sorted(clip_bank.keys())
                track_bank[f"{self.prefix}-tr{i}"] = VOXFactoryVocalTrack(
                    clip_bank=clip_bank,
                    clip_order=clip_order,
                    name=track.title,
                    mute=track.mute,
                    solo=track.solo,
                    pan=track.pan,
                )
        return track_bank, audio_data_bank

    def generate_notes(self, notes: list[Note], params: Params) -> dict[str, VOXFactoryVocalClip]:
        note_bank = {}
        note_order = []
        max_ticks = notes[-1].end_pos if notes else 0
        max_quarter = max_ticks / TICKS_IN_BEAT
        for i, note in enumerate(notes):
            note_bank[f"{self.prefix}-no{i}"] = self.generate_note(note, params)
            note_order.append(f"{self.prefix}-no{i}")
        clip_count = math.ceil(max_quarter / 32)
        clip_bank = {}
        for i in range(clip_count):
            clip_bank[f"{self.prefix}-cl{i}"] = VOXFactoryVocalClip(
                start_quarter=32 * i,
                offset_quarter=32 * i,
                length=32,
                note_bank=note_bank,
                note_order=note_order,
            )
        return clip_bank

    def generate_note(self, note: Note, params: Params) -> VOXFactoryNote:
        note_start_time = self.synchronizer.get_actual_secs_from_ticks(note.start_pos)
        return VOXFactoryNote(
            time=note_start_time,
            ticks=note.start_pos,
            duration_ticks=note.length,
            midi=note.key_number,
            name=note.lyric,
            syllable=note.pronunciation or DEFAULT_PHONEME,
            pitch_bends=self.generate_note_pitch(note, params.pitch),
        )

    def generate_note_pitch(self, note: Note, pitch: ParamCurve) -> list[float]:
        note_start_time = self.synchronizer.get_actual_secs_from_ticks(note.start_pos)
        note_end_time = self.synchronizer.get_actual_secs_from_ticks(note.end_pos)
        key_interval_dict = PiecewiseIntervalDict()
        secs_step = 1024 / 44100
        prev_secs = None
        prev_key: float = -1
        for point in pitch.points.root:
            if point.x - self.first_bar_length < note.start_pos:
                continue
            elif point.x - self.first_bar_length > note.end_pos:
                break
            if point.y == -100:
                prev_secs = None
                prev_key = 0
            else:
                secs = self.synchronizer.get_actual_secs_from_ticks(point.x - self.first_bar_length)
                key = point.y / 100
                if prev_secs is not None:
                    key_interval_dict[portion.openclosed(prev_secs, secs)] = functools.partial(
                        linear_interpolation,  # type: ignore[call-arg]
                        start=(prev_secs, prev_key - note.key_number),
                        end=(secs, key - note.key_number),
                    )
                else:
                    key_interval_dict[portion.singleton(secs)] = key - note.key_number
                prev_secs = secs
                prev_key = key
        pitch_bends = [
            key_interval_dict.get(secs, 0)
            for secs in more_itertools.numeric_range(note_start_time, note_end_time, secs_step)
        ]
        return pitch_bends if any(pitch_bends) else []



import sys
import zipfile
from importlib.machinery import ModuleSpec, SourcelessFileLoader
from importlib.util import module_from_spec, spec_from_file_location
from types import ModuleType
from typing import TYPE_CHECKING, cast

from libresvip.core.compat import Traversable

if TYPE_CHECKING:
    import pathlib
    from importlib.abc import Loader


class ZipLoader(SourcelessFileLoader):
    def __init__(self, zip_file: zipfile.ZipFile, file_path: str) -> None:
        self.zip_file = zip_file
        self.file_path = file_path

    def create_module(self, spec: ModuleSpec) -> ModuleType | None:
        return sys.modules.get(spec.name)

    def get_filename(self, name: str | None = None) -> str:
        return self.file_path

    def get_data(self, path: str) -> bytes:
        return self.zip_file.read(path)

    def exec_module(self, module: ModuleType) -> None:
        if compiled := super().get_code(module.__name__):
            exec(compiled, module.__dict__)


def load_module(name: str, plugin_path: Traversable) -> ModuleType:
    spec = None
    if (
        isinstance(plugin_path, zipfile.Path)
        and plugin_path.root.filename is not None
        and hasattr(plugin_path, "at")
    ):
        loader = ZipLoader(zip_file=plugin_path.root, file_path=plugin_path.at)
        spec = ModuleSpec(name, cast("Loader", loader), is_package=True, origin=plugin_path.at)
    else:
        spec = spec_from_file_location(
            name,
            cast("pathlib.Path", plugin_path),
            submodule_search_locations=[
                str(plugin_path)
                if plugin_path.is_dir()
                else str(plugin_path).removesuffix(plugin_path.name)
            ],
        )
    if spec is None or spec.loader is None:
        msg = f"Cannot load plugin from {plugin_path}"
        raise ImportError(msg)
    module = module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as e:
        msg = f"{e}: {plugin_path}"
        raise ImportError(msg) from e
    return module


def import_module(module_name: str, file_path: Traversable, reload: bool) -> ModuleType:
    """
    Import a module, trying either to find it as a single file or as a directory.

    .. note:: Isolated and provided to be reused, but not to be reimplemented !
    """
    if module_name not in sys.modules or reload:
        sys.modules[module_name] = load_module(module_name, file_path)

    return sys.modules[module_name]



import re
from typing import Any

from PySide6.QtCore import (
    QAbstractTableModel,
    QByteArray,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
    Slot,
)

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from libresvip.core.config import (
    LyricsReplacement,
    LyricsReplaceMode,
    settings,
)
from libresvip.extension.manager import plugin_manager


def _(text: str) -> str:
    return text


class PluginCadidatesTableModel(QAbstractTableModel):
    def __init__(self) -> None:
        super().__init__()
        self.column_names = [
            _("Applicable File Format"),
            _("Author"),
            _("Version"),
            "On/Off",
        ]
        self.plugin_candidates: list[dict[int, str]] = []
        self.reload_formats()

    def reload_formats(self) -> None:
        self.begin_remove_rows(QModelIndex(), 0, len(self.plugin_candidates) - 1)
        self.plugin_candidates.clear()
        self.end_remove_rows()
        self.begin_insert_rows(QModelIndex(), 0, len(plugin_manager._candidates) - 1)
        self.plugin_candidates = [
            {
                0: plugin.file_format,
                1: plugin.author,
                2: str(plugin.version),
                3: "checkbox-marked"
                if plugin.suffix not in settings.disabled_plugins
                else "checkbox-blank-outline",
            }
            for _path, plugin in plugin_manager._candidates
        ]
        self.end_insert_rows()

    def row_count(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:  # type: ignore[assignment]
        return len(self.plugin_candidates)

    def column_count(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:  # type: ignore[assignment]
        return len(self.column_names)

    def header_data(
        self,
        section: int,
        orientation: Qt.Orientation,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> str | None:
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if orientation == Qt.Orientation.Horizontal and 0 <= section < len(self.column_names):
            return self.column_names[section]
        return str(section + 1)

    def data(
        self,
        index: QModelIndex | QPersistentModelIndex,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> Any:
        if role == Qt.ItemDataRole.DisplayRole or (
            role == Qt.ItemDataRole.EditRole and index.column() == 3
        ):
            return self.plugin_candidates[index.row()][index.column()]
        return None

    def flags(self, index: QModelIndex | QPersistentModelIndex) -> Qt.ItemFlag:
        item_flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        if index.column() == 3:
            item_flags |= Qt.ItemFlag.ItemIsEditable
        return item_flags

    def role_names(self) -> dict[int, QByteArray]:
        return {
            Qt.ItemDataRole.DisplayRole: QByteArray(b"display"),
            Qt.ItemDataRole.EditRole: QByteArray(b"value"),
        }


class LyricReplacementRulesTableModel(QAbstractTableModel):
    def __init__(self, preset: str) -> None:
        super().__init__()
        self.preset = preset
        self.column_keys = [
            "mode",
            "pattern_prefix",
            "pattern_main",
            "pattern_suffix",
            "replacement",
            "flags",
        ]
        self.column_names = [
            _("Mode"),
            _("Prefix"),
            _("Pattern"),
            _("Suffix"),
            _("Replacement"),
            _("Case sensitive"),
            _("Actions"),
        ]

    def row_count(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:  # type: ignore[assignment]
        return len(settings.lyric_replace_rules[self.preset])

    def column_count(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:  # type: ignore[assignment]
        return len(self.column_names)

    def header_data(
        self,
        section: int,
        orientation: Qt.Orientation,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> str | None:
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if orientation == Qt.Orientation.Horizontal and 0 <= section < len(self.column_names):
            return self.column_names[section]
        return str(section + 1)

    def data(
        self,
        index: QModelIndex | QPersistentModelIndex,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> Any:
        column_index = index.column()
        if role in [
            Qt.ItemDataRole.DisplayRole,
            Qt.ItemDataRole.EditRole,
        ] and 0 <= column_index < len(self.column_keys):
            prop = getattr(
                settings.lyric_replace_rules[self.preset][index.row()],
                self.column_keys[column_index],
            )
            if isinstance(prop, str):
                return prop
            elif isinstance(prop, re.RegexFlag):
                return prop.name
            elif isinstance(prop, LyricsReplaceMode):
                return prop.value
        return ""

    def set_data(
        self,
        index: QModelIndex | QPersistentModelIndex,
        value: Any,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> bool:
        key = self.column_keys[index.column()]
        row = index.row()
        rule = settings.lyric_replace_rules[self.preset][row]
        if key == "mode":
            return False
        elif key == "flags":
            flags = getattr(
                re.RegexFlag,
                "IGNORECASE" if value == "IGNORECASE" else "UNICODE",
            )
            setattr(rule, key, flags)
            value = flags.name
        else:
            setattr(rule, key, value)
        return True

    def flags(self, index: QModelIndex | QPersistentModelIndex) -> Qt.ItemFlag:
        item_flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        column = index.column()
        if 0 < column < len(self.column_keys) and (column_key := self.column_keys[column]):
            editable = True
            if column_key in ["pattern_prefix", "pattern_suffix"]:
                row = index.row()
                if settings.lyric_replace_rules[self.preset][row].mode != LyricsReplaceMode.REGEX:
                    editable = False
            if editable:
                item_flags |= Qt.ItemFlag.ItemIsEditable
        return item_flags

    def role_names(self) -> dict[int, QByteArray]:
        return {
            Qt.ItemDataRole.DisplayRole: QByteArray(b"display"),
            Qt.ItemDataRole.EditRole: QByteArray(b"value"),
        }

    @Slot(str)
    def append(self, mode: str) -> None:
        settings.lyric_replace_rules[self.preset].append(
            LyricsReplacement(  # type: ignore[call-arg]
                mode=LyricsReplaceMode(mode), pattern_main="", replacement=""
            )
        )
        self.modelReset.emit()

    @Slot(int, int)
    def swap(self, old_index: int, new_index: int) -> None:
        if (
            not (0 <= old_index < len(settings.lyric_replace_rules[self.preset]))
            or not (0 <= new_index < len(settings.lyric_replace_rules[self.preset]))
            or (old_index + 1 != new_index)
        ):
            return
        self.begin_move_rows(QModelIndex(), old_index, old_index, QModelIndex(), new_index + 1)
        rule = settings.lyric_replace_rules[self.preset].pop(new_index)
        settings.lyric_replace_rules[self.preset].insert(old_index, rule)
        self.end_move_rows()

    @Slot(int)
    def delete(self, index: int) -> None:
        self.begin_remove_rows(QModelIndex(), index, index)
        settings.lyric_replace_rules[self.preset].pop(index)
        self.end_remove_rows()



import fractions

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import zoom_project

from .options import ProcessOptions


class ProjectZoomMiddleware(plugin_base.MiddlewareBase):
    def process(self, project: Project, options: ProcessOptions) -> Project:
        if (zoom_factor := float(fractions.Fraction(options.factor.value))) != 1.0:
            return zoom_project(project, zoom_factor)
        return project



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VocalShifterProjectData
from .options import InputOptions
from .vocalshifter_parser import VocalShifterParser


class VocalShifterConverter(plugin_base.ReadOnlyConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        vshp_proj = VocalShifterProjectData.parse_file(path)
        return VocalShifterParser(options).parse_project(vshp_proj)



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class OpenUtauEnglishPhonemizerCompatibility(Enum):
    NON_ARPA: Annotated[
        str,
        Field(
            title=_("Incompatible with ARPAsing-series Phonemizers"),
        ),
    ] = "non-arpa"
    ARPA: Annotated[
        str,
        Field(
            title=_("Compatible with ARPAsing-series Phonemizers"),
        ),
    ] = "arpa"


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    BaseModel,
):
    english_phonemizer_compatibility: OpenUtauEnglishPhonemizerCompatibility = Field(
        OpenUtauEnglishPhonemizerCompatibility.NON_ARPA,
        title=_("The way to handle english multisyllabic words"),
        description=_("Compatibility with ARPAsing-series Phonemizer"),
    )
    breath_lyrics: str = Field(
        "Asp AP",
        title=_("Breath lyrics"),
        description=_(
            "Special lyrics that will be recognized as breath notes only when immediately followed by a regular note, each separated by a space"
        ),
    )
    silence_lyrics: str = Field(
        "R SP",
        title=_("Silence lyrics"),
        description=_("Special lyrics that will be ignored, each separated by a space"),
    )


class OutputOptions(BaseModel):
    english_phonemizer_compatibility: OpenUtauEnglishPhonemizerCompatibility = Field(
        OpenUtauEnglishPhonemizerCompatibility.NON_ARPA,
        title=_("The way to handle english multisyllabic words"),
        description=_("Compatibility with ARPAsing-series Phonemizer"),
    )



from collections.abc import Callable
from typing import TypeVar

from more_itertools import locate, rlocate

T = TypeVar("T")


def find_index(obj_list: list[T], pred: Callable[[T], bool]) -> int:
    return next(locate(obj_list, pred), -1)


def find_last_index(obj_list: list[T], pred: Callable[[T], bool]) -> int:
    return next(rlocate(obj_list, pred), -1)


def binary_find_first(n: list[T], pred: Callable[[T], bool]) -> T | None:
    if not len(n):
        return None
    left, right = 0, len(n) - 1
    while right > left:
        middle = (left + right) // 2
        if pred(n[middle]):
            right = middle
        else:
            left = middle + 1
    return n[right] if pred(n[right]) else None


def binary_find_last(n: list[T], pred: Callable[[T], bool]) -> T | None:
    if not len(n):
        return None
    left, right = 0, len(n) - 1
    while right > left:
        middle = (left + right + 1) // 2
        if pred(n[middle]):
            left = middle
        else:
            right = middle - 1
    return n[left] if pred(n[left]) else None



import collections
import dataclasses
import itertools
import math
import operator

import mido_fix as mido
import more_itertools

from libresvip.core.constants import (
    DEFAULT_PHONEME,
    DEFAULT_PITCH_BEND_SENSITIVITY,
    PITCH_MAX_VALUE,
    TICKS_IN_BEAT,
)
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    Note,
    ParamCurve,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.music_math import ratio_to_db
from libresvip.utils.text import LATIN_ALPHABET
from libresvip.utils.translation import gettext_lazy as _

from .constants import (
    EXPRESSION_CONSTANT,
    VELOCITY_CONSTANT,
    ControlChange,
)
from .note_overlap import has_overlap
from .options import InputOptions, MultiChannelOption


def cc11_to_db_change(value: float) -> float:
    return ratio_to_db((value / 127) ** EXPRESSION_CONSTANT + 1e-6)


def velocity_to_db_change(value: float) -> float:
    return ratio_to_db((value / 127) ** VELOCITY_CONSTANT + 1e-6)


@dataclasses.dataclass
class MidiParser:
    options: InputOptions
    ticks_per_beat: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    selected_channels: list[int] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        if self.options.multi_channel == MultiChannelOption.FIRST:
            self.selected_channels = [0]
        elif self.options.multi_channel == MultiChannelOption.CUSTOM:
            for exp in self.options.channels.split(","):
                if exp:
                    start, sep, end = exp.partition("-")
                    if start.isdigit() and not start.startswith("0"):
                        if not len(end):
                            self.selected_channels.append(int(start) - 1)
                        elif end.isdigit() and not end.startswith("0"):
                            self.selected_channels.extend(range(int(start) - 1, int(end)))

    @property
    def tick_rate(self) -> float:
        return TICKS_IN_BEAT / self.ticks_per_beat

    def parse_project(self, mido_obj: mido.MidiFile) -> Project:
        self.ticks_per_beat = mido_obj.ticks_per_beat
        self._convert_delta_to_cumulative(mido_obj.tracks)
        time_signature_list = []
        if len(mido_obj.tracks):
            master_track = mido_obj.tracks[0]
            time_signature_list.extend(self.parse_time_signatures(master_track))
        song_tempo_list = self.parse_tempo(mido_obj.tracks)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=time_signature_list,
            track_list=self.parse_tracks(mido_obj.tracks),
        )

    @staticmethod
    def _convert_delta_to_cumulative(tracks: list[mido.MidiTrack]) -> None:
        for track in tracks:
            tick = 0
            for event in track:
                event.time += tick
                tick = event.time

    def parse_time_signatures(self, master_track: mido.MidiTrack) -> list[TimeSignature]:
        # no default
        time_signature_changes: list[TimeSignature] = []

        # traversing
        if self.options.import_time_signatures:
            prev_ticks = 0
            measure = 0
            for event in master_track:
                if event.type == "time_signature":
                    tick = event.time
                    if not time_signature_changes:
                        tick_in_full_note = 4 * self.ticks_per_beat
                    else:
                        tick_in_full_note = round(
                            time_signature_changes[-1].bar_length(self.ticks_per_beat)
                        )
                    measure += (tick - prev_ticks) / tick_in_full_note
                    ts_obj = TimeSignature(
                        bar_index=math.floor(measure),
                        numerator=event.numerator,
                        denominator=event.denominator,
                    )
                    time_signature_changes.append(ts_obj)
                    prev_ticks = tick
        if not time_signature_changes or time_signature_changes[0].bar_index > 0:
            time_signature_changes.insert(0, TimeSignature(bar_index=0, numerator=4, denominator=4))
        self.first_bar_length = round(time_signature_changes[0].bar_length())
        return time_signature_changes

    def parse_tempo(self, tracks: list[mido.MidiTrack]) -> list[SongTempo]:
        tempos: list[SongTempo] = []

        # traversing
        for track in tracks:
            for event in track:
                if event.type == "set_tempo":
                    # convert tempo to BPM
                    tempo = round(mido.tempo2bpm(event.tempo), 3)
                    tick = round(event.time * self.tick_rate)
                    last_tempo = tempos[-1].bpm if tempos else None
                    if tempo != last_tempo:
                        tempos.append(SongTempo(position=tick, bpm=tempo))
        if not tempos:
            # default bpm
            show_warning(_("No tempo labels found in the imported project."))
            tempos.append(SongTempo(position=0, bpm=self.options.default_bpm))
        else:
            tempos.sort(key=operator.attrgetter("position"))
        return tempos

    def parse_track(self, track_idx: int, track: mido.MidiTrack) -> list[SingingTrack]:
        tracks = []
        event_buckets = more_itertools.bucket(
            (event for event in track if hasattr(event, "channel")),
            key=operator.attrgetter("channel"),
        )
        lyrics: dict[int, str] = collections.defaultdict(lambda: DEFAULT_PHONEME)
        if self.options.import_lyrics:
            for event in track:
                if event.type == "lyrics":
                    lyrics[event.time] = event.text
        for channel in event_buckets:
            if self.selected_channels and channel not in self.selected_channels:
                continue
            last_note_on = collections.defaultdict(list)
            pitchbend_range_changed: dict[int, list[int]] = collections.defaultdict(list)
            track_name = None
            notes = []
            rel_pitch_points = []
            expression = ParamCurve()
            pitch_bend_sensitivity = DEFAULT_PITCH_BEND_SENSITIVITY
            volume_base = 0.0
            for event in event_buckets[channel]:
                # Look for track name events
                if event.type == "track_name":
                    # Set the track name for the current track
                    track_name = event.name
                elif event.type == "note_on" and event.velocity > 0:
                    # Store this as the last note-on location
                    rel_pitch_points.append(Point(round(event.time * self.tick_rate), 0))
                    last_note_on[event.note].append(event.time)
                elif event.type == "note_off" or (event.type == "note_on" and event.velocity == 0):
                    # Check that a note-on exists (ignore spurious note-offs)
                    key = event.note
                    if key in last_note_on:
                        # Get the start/stop times and velocity of every note
                        # which was turned on with this instrument/drum/pitch.
                        # One note-off may close multiple note-on events from
                        # previous ticks. In case there's a note-off and then
                        # note-on at the same tick we keep the open note from
                        # this tick.
                        end_tick = event.time
                        open_notes = last_note_on[key]

                        notes_to_close = [
                            start_tick for start_tick in open_notes if start_tick != end_tick
                        ]
                        notes_to_keep = [
                            start_tick for start_tick in open_notes if start_tick == end_tick
                        ]

                        for start_tick in notes_to_close:
                            # Create the note event
                            note = Note(
                                key_number=event.note,
                                start_pos=round(start_tick * self.tick_rate),
                                length=round((end_tick - start_tick) * self.tick_rate),
                            )
                            lyric = lyrics[start_tick]
                            if LATIN_ALPHABET.search(lyric) is not None:
                                note.pronunciation = lyric
                            note.lyric = lyric
                            notes.append(note)
                        if notes_to_close and notes_to_keep:
                            # Note-on on the same tick but we already closed
                            # some previous notes -> it will continue, keep it.
                            last_note_on[key] = notes_to_keep
                        else:
                            # Remove the last note on for this instrument
                            del last_note_on[key]
                elif event.type == "pitchwheel" and self.options.import_pitch:
                    # Create pitch bend class instance
                    rel_pitch_points.append(
                        Point(
                            round(event.time * self.tick_rate),
                            round(
                                pitch_bend_sensitivity
                                * event.pitch
                                / (PITCH_MAX_VALUE if event.pitch > 0 else (PITCH_MAX_VALUE + 1))
                            ),
                        )
                    )
                elif event.type == "control_change":
                    if self.options.import_pitch:
                        if (
                            event.is_cc(ControlChange.DATA_ENTRY)
                            and len(pitchbend_range_changed[event.time]) >= 2
                        ):
                            pitch_bend_sensitivity = event.value
                        elif (event.is_cc(ControlChange.RPN_MSB) and event.value == 0) or (
                            event.is_cc(ControlChange.RPN_LSB) and event.value == 0
                        ):
                            pitchbend_range_changed[event.time].append(event.value)
                    if self.options.import_volume:
                        if event.is_cc(ControlChange.EXPRESSION) and event.value:
                            expression.points.append(
                                Point(
                                    round(event.time * self.tick_rate),
                                    round(volume_base + cc11_to_db_change(event.value)),
                                )
                            )
                        elif event.is_cc(ControlChange.VOLUME) and event.value:
                            volume_base = velocity_to_db_change(event.value)
            if has_overlap(notes):
                msg = _("Overlapping notes in track {}").format(track_idx)
                show_warning(msg)
            edited_params = Params(volume=expression)
            if self.options.import_pitch:
                pitch_simulator = PitchSimulator(
                    synchronizer=self.synchronizer,
                    portamento=PortamentoPitch.no_portamento(),
                    note_list=notes,
                )
                rel_pitch_points.sort(key=operator.attrgetter("x"))
                edited_params.pitch = RelativePitchCurve(self.first_bar_length).to_absolute(
                    rel_pitch_points, pitch_simulator
                )
            if notes:
                tracks.append(
                    SingingTrack(
                        title=track_name or f"Track {track_idx + 1} ({channel})",
                        note_list=notes,
                        edited_params=edited_params,
                    )
                )
        return tracks

    def parse_tracks(self, midi_tracks: list[mido.MidiTrack]) -> list[SingingTrack]:
        return [
            *itertools.chain.from_iterable(
                (self.parse_track(track_idx, track) for track_idx, track in enumerate(midi_tracks)),
            )
        ]



import dataclasses
import hashlib
import io
import pathlib

import more_itertools

from libresvip.core.constants import DEFAULT_CHINESE_LYRIC, TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import InstrumentalTrack, Note, Params, Project, SingingTrack, SongTempo
from libresvip.utils.audio import audio_track_info

from .enums import PocketSingerLyricsLanguage
from .model import (
    PocketSingerBgmInfo,
    PocketSingerBgmTrack,
    PocketSingerMetadata,
    PocketSingerNote,
    PocketSingerPitchBend,
    PocketSingerProject,
    PocketSingerSongInfo,
    PocketSingerTrack,
)
from .options import OutputOptions


@dataclasses.dataclass
class PocketSingerGenerator:
    options: OutputOptions
    buffer: io.BytesIO = dataclasses.field(default_factory=io.BytesIO)
    audio_paths: dict[str, pathlib.Path] = dataclasses.field(default_factory=dict)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> PocketSingerMetadata:
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = TICKS_IN_BEAT * 4
        song_info = self.generate_song_info(project.song_tempo_list[0])
        ps_project = PocketSingerProject(song_info=song_info)
        if (
            first_instrumental_track := next(
                (track for track in project.track_list if isinstance(track, InstrumentalTrack)),
                None,
            )
        ) and (bgm_info := self.generate_bgm_info(first_instrumental_track)):
            ps_project.bgm_info = bgm_info
        for track in project.track_list:
            if isinstance(track, SingingTrack) and (ps_track := self.generate_track(track)):
                ps_project.tracks.append(ps_track)
        self.buffer.write(
            ps_project.model_dump_json(by_alias=True, exclude_none=True).encode("utf-8")
        )
        return PocketSingerMetadata(ace_file_name="ps_structure.ps")

    def generate_song_info(self, tempo: SongTempo) -> PocketSingerSongInfo:
        return PocketSingerSongInfo(
            start=0,
            first_beat_offset=0.0,
            scale=[
                80,
                79,
                77,
                76,
                75,
                74,
                73,
                72,
                71,
                70,
                69,
                68,
                67,
                66,
                65,
                64,
                63,
                62,
                61,
                60,
                59,
                58,
                57,
                56,
                55,
                54,
                53,
                52,
                51,
                50,
                49,
                48,
                43,
            ],
            key="C",
            segment_of_beat=2,
            operate_scale=[72, 71, 69, 67, 65, 64, 62, 60, 59, 57, 55],
            bpm=tempo.bpm,
            name=self.options.title,
            duration=0.0,
            beat_of_bar=4,
        )

    def generate_bgm_info(
        self, instrumental_track: InstrumentalTrack
    ) -> PocketSingerBgmInfo | None:
        audio_path = pathlib.Path(instrumental_track.audio_file_path)
        file_type = audio_path.suffix[1:].lower()
        if file_type not in ("wav", "mp3", "flac", "aac", "m4a"):
            return None
        if not (track_info := audio_track_info(audio_path)):
            return None
        audio_content = audio_path.read_bytes()
        self.audio_paths[audio_path.name] = audio_path
        audio_position = self.synchronizer.get_actual_secs_from_ticks(instrumental_track.offset)
        return PocketSingerBgmInfo(
            mute=instrumental_track.mute,
            solo=instrumental_track.solo,
            tracks=[
                PocketSingerBgmTrack(
                    file_name=audio_path.stem,
                    file_type=file_type,
                    file_md5=hashlib.md5(audio_content).hexdigest(),
                    position=audio_position,
                    start_time=audio_position,
                    end_time=audio_position + track_info.duration / 1000,
                )
            ],
        )

    def generate_track(self, track: SingingTrack) -> PocketSingerTrack | None:
        ps_track = PocketSingerTrack(
            language=self.options.lyric_language,
            mute=track.mute,
            solo=track.solo,
            pan=track.pan,
            sound_effect=0,
            singer_volume=1,
        )
        if patched_notes := self.patch_notes(track.note_list):
            ps_notes, ps_track.lyric = self.generate_notes(patched_notes, track.edited_params)
            ps_track.notes.extend(ps_notes)
            return ps_track

    def patch_notes(self, notes: list[Note]) -> list[Note]:
        prev_note = None
        patched_notes = []
        for note in notes:
            if not note.lyric:
                continue
            elif note.lyric in ["-", "+"] and prev_note is not None:
                if note.start_pos == prev_note.end_pos:
                    note.key_number = prev_note.key_number
                    if note.lyric == "-":
                        prev_note.length += note.length
                        continue
                    else:
                        patched_notes.append(note)
                else:
                    prev_note = None
                    continue
            else:
                patched_notes.append(note)
            prev_note = note
        return patched_notes

    def generate_notes(
        self, notes: list[Note], params: Params
    ) -> tuple[list[PocketSingerNote], str]:
        ps_notes = []
        lyrics = []
        for note_group in more_itertools.split_when(
            notes,
            lambda prev_note, note: note.lyric != "+",
        ):
            group_lyric = note_group[0].lyric
            for i, note in enumerate(note_group):
                ps_note = PocketSingerNote(
                    language=self.options.lyric_language,
                    grapheme_count=len(note_group),
                    grapheme_index=i,
                    grapheme=group_lyric,
                    pitch=note.key_number,
                    start_time=self.synchronizer.get_actual_secs_from_ticks(note.start_pos),
                    end_time=self.synchronizer.get_actual_secs_from_ticks(note.end_pos),
                )
                if (
                    self.options.lyric_language != PocketSingerLyricsLanguage.ENGLISH
                    and len(group_lyric) == 1
                ):
                    lyrics.append(group_lyric)
                else:
                    lyrics.append(DEFAULT_CHINESE_LYRIC)
                if note.edited_phones is not None:
                    ps_note.consonant_time_head = [note.edited_phones.head_length_in_secs]
                for point in params.pitch.points.root:
                    if note.start_pos <= point.x - self.first_bar_length <= note.end_pos:
                        ps_note.pitch_bends.append(
                            PocketSingerPitchBend(
                                time=self.synchronizer.get_actual_secs_from_ticks(
                                    point.x - self.first_bar_length
                                ),
                                pitch=(point.y / 100 - note.key_number) if point.y != -100 else 0,
                            )
                        )
                ps_notes.append(ps_note)
        return ps_notes, "".join(lyrics)



import enum
import itertools
from typing import TYPE_CHECKING, Annotated, Literal, cast

import more_itertools
from pydantic import (
    Field,
    SerializationInfo,
    ValidationInfo,
    computed_field,
    field_serializer,
    field_validator,
)

from libresvip.model.base import BaseModel

if TYPE_CHECKING:
    from collections.abc import Iterable


class AISTrackType(enum.IntEnum):
    TRACK_SING_VOICE = 0
    TRACK_AUDIO = 1
    TRACK_MIDI = 2


class AISNote(BaseModel):
    start: int = Field(alias="s")
    length: int = Field(alias="l")
    midi_no: int = Field(alias="m")
    lyric: str | None = Field(alias="ly")
    pinyin: str | None = Field(alias="py")
    vel: int | None = 50
    triple: bool | None = Field(False, alias="tri")
    pit: list[float] = Field(default_factory=list)
    bc: int | None = 0
    bj: int | None = 0
    bq: int | None = 0

    @field_validator("pit", mode="before")
    @classmethod
    def validate_pit(
        cls, value: str | list[float | str], _info: ValidationInfo
    ) -> list[float] | None:
        if value is None:
            return None
        value_list = value.split() if isinstance(value, str) else value
        pit_list = []
        for x in cast("Iterable[float | str]", value_list):
            if isinstance(x, str) and "x" in x:
                x, _, repeat_times = x.partition("x")
                pit_list.extend([float(x)] * int(repeat_times))
            elif isinstance(x, str):
                pit_list.append(float(x))
            else:
                pit_list.append(x)
        return pit_list

    @field_serializer("pit", when_used="json-unless-none")
    @classmethod
    def serialize_pit(cls, value: list[float] | None, _info: SerializationInfo) -> str:
        if value is None:
            return "0x500"
        pit_str = ""
        for key, group in itertools.groupby(value):
            group_length = more_itertools.ilen(group)
            pit_str += (
                f"{round(key, 2)}x{group_length} " if group_length > 1 else f"{round(key, 2)} "
            )
        return pit_str.strip()


class AISBasePattern(BaseModel):
    uid: int | None = None
    start: int = Field(alias="s")
    length: int | None = Field(alias="l")


class AISSingVoicePattern(AISBasePattern):
    notes: list[AISNote] = Field(default_factory=list, alias="n")


class AISAudioPattern(AISBasePattern):
    path_audio: str | None = Field(alias="pa")
    path_wave: str | None = Field(alias="pw")
    n_channel: int | None = 2
    len_sec: int | None = 0


class AISBaseTrack(BaseModel):
    idx: int | None = Field(alias="i")
    solo: bool | None = Field(False, alias="s")
    mute: bool | None = Field(False, alias="m")
    volume: float | None = Field(0, alias="v")
    name: str | None = Field(alias="n")


class AISSingVoiceTrack(AISBaseTrack):
    track_type: Literal[AISTrackType.TRACK_SING_VOICE] = Field(
        AISTrackType.TRACK_SING_VOICE, alias="t"
    )
    singer_namecn: str | None = Field(alias="sn")
    singer_nameen: str | None = Field("", alias="se")
    singer_head_path: str | None = Field("", alias="sh")
    items: list[AISSingVoicePattern] = Field(alias="im", default_factory=list)


class AISAudioTrack(AISBaseTrack):
    track_type: Literal[AISTrackType.TRACK_AUDIO] = Field(AISTrackType.TRACK_AUDIO, alias="t")
    items: list[AISAudioPattern] = Field(alias="im", default_factory=list)


class AISMidiTrack(AISBaseTrack):
    track_type: Literal[AISTrackType.TRACK_MIDI] = Field(AISTrackType.TRACK_MIDI, alias="t")


AISTrack = Annotated[
    AISSingVoiceTrack | AISAudioTrack | AISMidiTrack,
    Field(discriminator="track_type"),
]


class AISTimeSignature(BaseModel):
    beat_zi: int = 4
    beat_mu: int = 4
    start_bar: int = 0

    @computed_field(alias="str")
    def str_value(self) -> str:
        return f"{self.beat_zi}/{self.beat_mu}"


class AISTempo(BaseModel):
    tempo_float: float | None = None
    start_128: int
    start_bar: int
    start_beat_in_bar: int | None = None


class AISProjectBody(BaseModel):
    tracks: list[AISTrack] = Field(default_factory=list)

    @computed_field
    def num_track(self) -> int:
        return len(self.tracks)


class AISProjectHead(BaseModel):
    tempo: list[AISTempo] = Field(default_factory=list)
    signature: list[AISTimeSignature] = Field(default_factory=list)
    flags: int | None = -256
    flage: int | None = -128
    time: int | None = None
    bar: int | None = None



import pathlib

from jinja2 import Template

from .model import UTAUProject

UST_TEMPLATE = Template(
    """\
[#VERSION]
UST Version={{ ust_project.ust_version }}{% if ust_project.charset %}
Charset={{ ust_project.charset }}{% endif %}
[#SETTING]{% if ust_project.tempo | round(2) %}
Tempo={{ ust_project.tempo }}{% endif %}{% if ust_project.track_count %}
Tracks={{ ust_project.track_count }}{% endif %}{% if ust_project.project_name %}
Project={{ ust_project.project_name }}{% endif %}{% if ust_project.voice_dir %}
VoiceDir={{ ust_project.voice_dir }}{% endif %}{% if ust_project.out_file %}
OutFile={{ ust_project.out_file }}{% endif %}{% if ust_project.cache_dir %}
CacheDir={{ ust_project.cache_dir }}{% endif %}{% if ust_project.tool1 %}
Tool1={{ ust_project.tool1 }}{% endif %}{% if ust_project.tool2 %}
Tool2={{ ust_project.tool2 }}{% endif %}{% if ust_project.pitch_mode2 %}
Mode2={{ ust_project.pitch_mode2 | round }}{% endif %}{% if ust_project.autoren %}
Autoren={{ ust_project.autoren }}{% endif %}{% if ust_project.flags %}
Flags={{ ust_project.flags }}{% endif %}{% if ust_project.track | length > 0 %}{% for note in ust_project.track[0].notes %}
[#{{ note.note_type }}]
Length={{ note.length }}
Lyric={{ note.lyric }}
NoteNum={{ note.note_num }}{% if note.pre_utterance %}
PreUtterance={{ note.pre_utterance }}{% endif %}{% if note.voice_overlap %}
VoiceOverlap={{ note.voice_overlap | round(2) }}{% endif %}{% if note.intensity %}
Intensity={{ note.intensity }}{% endif %}{% if note.modulation %}
Modulation={{ note.modulation }}{% endif %}{% if note.start_point %}
StartPoint={{ note.start_point | round(2) }}{% endif %}{% if note.envelope %}
Envelope={{ note.envelope.p1 }},{{ note.envelope.p2 }},{{ note.envelope.p3 }},{{ note.envelope.v1 }},{{ note.envelope.v2 }},{{ note.envelope.v3 }},{{ note.envelope.v4 }}\
{% if note.envelope.v5 %}\
,%,{{ note.envelope.p4 }},{{ note.envelope.p5 }},{{ note.envelope.v5 }}{% elif note.envelope.p5 %}\
,%,{{ note.envelope.p4 }},{{ note.envelope.p5 }}{% elif note.envelope.p4 %}\
,,{{ note.envelope.p4 }}{% endif %}{% endif %}{% if note.tempo %}
Tempo={{ note.tempo | round(2) }}{% endif %}{% if note.velocity %}
Velocity={{ note.velocity | round(2) }}{% endif %}{% if note.label %}
Label={{ note.label }}{% endif %}{% if note.flags %}
Flags={{ note.flags }}{% endif %}{% if note.pitchbend_type %}
PBType={{ note.pitchbend_type }}{% endif %}{% if note.pitchbend_start %}
PBStart={{ note.pitchbend_start }}{% endif %}{% if note.pitch_bend_points|length > 0 %}
PitchBend={{ note.pitch_bend_points|join(',') }}{% endif %}{% if note.pbs|length > 0 %}
PBS={{ note.pbs | join(';') }}{% endif %}{% if note.pbw|length > 0 %}
PBW={{ note.pbw | join(',') }}{% endif %}{% if note.pby|length > 0 %}
PBY={{ note.pby | join(',') }}{% endif %}{% if note.pbm|length > 0 %}
PBM={{ note.pbm | join(',') }}{% endif %}{% if note.vbr %}
VBR={{ note.vbr.length }},{{ note.vbr.period }},{% if note.vbr.depth is defined %}\
{{ note.vbr.depth }}{% endif %},{% if note.vbr.fade_in is defined %}\
{{ note.vbr.fade_in }}{% endif %},{% if note.vbr.fade_out is defined %}\
{{ note.vbr.fade_out }}{% endif %},{% if note.vbr.phase_shift is defined %}\
{{ note.vbr.phase_shift }}{% endif %},{% if note.vbr.shift is defined %}\
{{ note.vbr.shift }}{% endif %}{% endif %}{% endfor %}
[#TRACKEND]{% endif %}
"""
)


def render_ust(
    ust_project: UTAUProject,
    output_path: pathlib.Path,
    encoding: str = "utf-8",
) -> None:
    output_path.write_bytes(UST_TEMPLATE.render(ust_project=ust_project).encode(encoding))



import enum
import io
import pathlib
import traceback
from importlib.resources import as_file
from typing import get_args, get_type_hints

import flet as ft
import flet_permission_handler as fph
import more_itertools
from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from pydantic_extra_types.color import Color
from upath import UPath

import libresvip
from libresvip.core.compat import ZipFile
from libresvip.core.config import settings
from libresvip.core.constants import res_dir
from libresvip.core.warning_types import CatchWarnings
from libresvip.extension.manager import get_translation, middleware_manager, plugin_manager
from libresvip.model.base import BaseComplexModel, Project
from libresvip.utils import translation
from libresvip.utils.translation import gettext_lazy as _


def main(page: ft.Page) -> None:
    page.title = "LibreSVIP"
    page.scroll = ft.ScrollMode.ADAPTIVE
    page.window.width = 480
    page.window.height = 720
    page.window.title_bar_hidden = True
    page.window.title_bar_buttons_hidden = True

    with as_file(res_dir / "libresvip.ico") as icon:
        page.window.icon = str(icon)

    if not page.client_storage.contains_key("dark_mode"):
        page.client_storage.set("dark_mode", "System")
    page.theme_mode = ft.ThemeMode((page.client_storage.get("dark_mode") or "System").lower())

    def change_theme(dark_mode: str) -> None:
        page.theme_mode = ft.ThemeMode(dark_mode.lower())
        page.client_storage.set("dark_mode", dark_mode)
        page.update()

    if not (page.client_storage.contains_key("language")):
        page.client_storage.set("language", "en_US")
    translation.singleton_translation = get_translation(page.client_storage.get("language"))

    def change_language(lang: str) -> None:
        page.client_storage.set("language", lang)
        translation.singleton_translation = get_translation(lang)
        page.go(f"/?lang={lang}")

    if not (page.client_storage.contains_key("save_folder")):
        page.client_storage.set("save_folder", ".")
    save_folder_text_field = ft.Ref[ft.TextField]()
    temp_path = UPath("memory:/")

    task_list_view = ft.Ref[ft.ListView]()
    input_select = ft.Ref[ft.Dropdown]()
    output_select = ft.Ref[ft.Dropdown]()
    conversion_mode_select = ft.Ref[ft.Dropdown]()
    input_options = ft.Ref[ft.ResponsiveRow]()
    output_options = ft.Ref[ft.ResponsiveRow]()
    middleware_options = {
        middleware_id: ft.Ref[ft.ExpansionPanel]()
        for middleware_id in middleware_manager.plugin_registry
    }

    def build_options(option_class: BaseModel) -> list[ft.Control]:
        fields = []
        for option_key, field_info in option_class.model_fields.items():
            default_value = None if field_info.default is PydanticUndefined else field_info.default
            if field_info.annotation is None:
                continue
            elif issubclass(field_info.annotation, enum.Enum):
                default_value = default_value.value if default_value is not None else None
                annotations = get_type_hints(
                    field_info.annotation,
                    include_extras=True,
                )
                choices = []
                for enum_item in field_info.annotation:
                    if enum_item.name in annotations:
                        annotated_args = list(
                            get_args(annotations[enum_item.name]),
                        )
                        if len(annotated_args) >= 2:
                            enum_field = annotated_args[1]
                        else:
                            continue
                        choices.append(
                            ft.dropdown.Option(
                                enum_item.value,
                                _(enum_field.title),
                                content=ft.Text(
                                    _(enum_field.title),
                                    tooltip=ft.Tooltip(
                                        _(enum_field.description or ""),
                                        trigger_mode=ft.TooltipTriggerMode.TAP,
                                    ),
                                ),
                            )
                        )
                fields.append(
                    ft.Dropdown(
                        options=choices,
                        label=_(field_info.title or ""),
                        value=default_value,
                        data=option_key,
                        col=10 if field_info.description is not None else 12,
                    )
                )
            elif issubclass(field_info.annotation, bool):
                fields.append(
                    ft.Switch(
                        label=_(field_info.title or ""),
                        value=default_value,
                        data=option_key,
                        col=10 if field_info.description is not None else 12,
                    )
                )
            elif issubclass(field_info.annotation, int):
                fields.append(
                    ft.TextField(
                        label=_(field_info.title or ""),
                        value=default_value,
                        data=option_key,
                        input_filter=ft.NumbersOnlyInputFilter(),
                        col=10 if field_info.description is not None else 12,
                    )
                )
            elif issubclass(field_info.annotation, float):
                fields.append(
                    ft.TextField(
                        label=_(field_info.title or ""),
                        value=default_value,
                        data=option_key,
                        input_filter=ft.InputFilter(regex_string=r"^-?\d+(\.\d*)?$"),
                        col=10 if field_info.description is not None else 12,
                    )
                )
            elif issubclass(field_info.annotation, str | BaseComplexModel | Color):
                if issubclass(field_info.annotation, BaseComplexModel):
                    default_value = field_info.annotation.default_repr()
                fields.append(
                    ft.TextField(
                        label=_(field_info.title or ""),
                        value=default_value,
                        data=option_key,
                        col=10 if field_info.description is not None else 12,
                    )
                )
            else:
                continue
            if field_info.description:
                fields.append(
                    ft.IconButton(
                        ft.Icons.HELP_OUTLINE_OUTLINED,
                        mouse_cursor=ft.MouseCursor.HELP,
                        tooltip=ft.Tooltip(_(field_info.description)),
                        col=2,
                    )
                )
        return fields

    def build_input_options(value: str | None) -> list[ft.Control]:
        if value in plugin_manager.plugin_registry:
            input_plugin = plugin_manager.plugin_registry[value]
            if (
                input_plugin.plugin_object is not None
                and (
                    input_option_cls := get_type_hints(input_plugin.plugin_object.load).get(
                        "options",
                    )
                )
                is not None
            ):
                return build_options(input_option_cls)
        return []

    def build_middleware_options(value: str) -> list[ft.Control]:
        if value in middleware_manager.plugin_registry:
            middleware = middleware_manager.plugin_registry[value]
            if (
                middleware.plugin_object is not None
                and (
                    middleware_option_cls := get_type_hints(middleware.plugin_object.process).get(
                        "options"
                    )
                )
                is not None
            ):
                return build_options(middleware_option_cls)
        return []

    def build_output_options(value: str | None) -> list[ft.Control]:
        if value in plugin_manager.plugin_registry:
            output_plugin = plugin_manager.plugin_registry[value]
            if (
                output_plugin.plugin_object is not None
                and (
                    output_option_cls := get_type_hints(output_plugin.plugin_object.dump).get(
                        "options",
                    )
                )
                is not None
            ):
                return build_options(output_option_cls)
        return []

    def set_last_input_format(value: str | None) -> None:
        if input_select.current.value != value:
            input_select.current.value = value
        last_input_format = page.client_storage.get("last_input_format")
        if last_input_format != value:
            input_options.current.controls = build_input_options(value)
            input_options.current.update()
            reset_tasks_on_input_change = page.client_storage.get("reset_tasks_on_input_change")
            if reset_tasks_on_input_change:
                task_list_view.current.controls.clear()
                task_list_view.current.update()
            page.client_storage.set("last_input_format", value)

    def set_last_output_format(value: str | None) -> None:
        if output_select.current.value != value:
            output_select.current.value = value
        last_output_format = page.client_storage.get("last_output_format")
        if last_output_format != value:
            output_options.current.controls = build_output_options(value)
            output_options.current.update()
            page.client_storage.set("last_output_format", value)

    def show_task_log(e: ft.ControlEvent) -> None:
        list_tile = e.control.parent.parent
        if list_tile.data.get("log_text"):

            def copy_log_text(e: ft.ControlEvent) -> None:
                page.set_clipboard(list_tile.data["log_text"])
                banner: ft.Banner = ft.Banner(
                    ft.Text(_("Copied")),
                    leading=ft.Icon(ft.Icons.CHECK_CIRCLE_OUTLINE),
                    actions=[ft.TextButton(text=_("OK"), on_click=lambda _: page.close(banner))],
                )
                page.open(banner)

            page.views.append(
                ft.View(
                    "/task_log",
                    appbar=ft.AppBar(
                        title=ft.Text(_("Task log")),
                        bgcolor=ft.Colors.SURFACE,
                        leading=ft.IconButton(
                            ft.Icons.ARROW_BACK_OUTLINED,
                            tooltip=_("Back"),
                            on_click=lambda _: view_pop(None),
                        ),
                    ),
                    controls=[
                        ft.Column(
                            [
                                ft.Row(
                                    [
                                        ft.ElevatedButton(
                                            text=_("Copy to clipboard"),
                                            on_click=copy_log_text,
                                        )
                                    ],
                                    alignment=ft.MainAxisAlignment.END,
                                ),
                                ft.TextField(
                                    value=list_tile.data["log_text"],
                                    multiline=True,
                                    max_lines=24,
                                    autofocus=True,
                                ),
                            ],
                        )
                    ],
                )
            )
            page.update()

    def on_files_selected(e: ft.FilePickerResultEvent) -> None:
        if e.files:
            auto_detect_input_format = page.client_storage.get("auto_detect_input_format")
            if page.web:
                uf = [
                    ft.FilePickerUploadFile(
                        f.name,
                        upload_url=page.get_upload_url(f.name, 600),
                    )
                    for f in e.files
                ]
                file_picker.upload(uf)
                return
            for file in e.files:
                last_input_format = page.client_storage.get("last_input_format")
                file_path = pathlib.Path(file.path)
                suffix = file_path.suffix.lower().removeprefix(".")
                if (
                    suffix != last_input_format
                    and auto_detect_input_format
                    and suffix in plugin_manager.plugin_registry
                ):
                    set_last_input_format(suffix)
                task_list_view.current.controls.append(
                    ft.ListTile(
                        leading=ft.Stack(
                            [
                                ft.Icon(
                                    ft.Icons.ACCESS_TIME_FILLED_OUTLINED,
                                    color=ft.colors.GREY_400,
                                ),
                                ft.ProgressRing(visible=False),
                            ]
                        ),
                        title=ft.Text(file.name),
                        subtitle=ft.Text(file_path.stem),
                        trailing=ft.PopupMenuButton(
                            items=[
                                ft.PopupMenuItem(
                                    icon=ft.icons.REMOVE_RED_EYE_OUTLINED,
                                    text=_("View Log"),
                                    on_click=show_task_log,
                                ),
                                ft.PopupMenuItem(
                                    icon=ft.icons.EDIT,
                                    text=_("Rename"),
                                    on_click=open_rename_dialog,
                                ),
                                ft.PopupMenuItem(
                                    icon=ft.icons.DELETE_OUTLINE,
                                    text=_("Remove"),
                                    on_click=remove_task,
                                ),
                            ],
                            tooltip=_("Actions"),
                        ),
                        data={"path": file_path, "log_text": ""},
                    )
                )
            page.update()
        if e.path:
            page.client_storage.set("save_folder", e.path)
            if save_folder_text_field.current is not None:
                save_folder_text_field.current.value = e.path
                save_folder_text_field.current.update()

    if not (page.client_storage.contains_key("auto_detect_input_format")):
        page.client_storage.set("auto_detect_input_format", True)

    def change_auto_detect_input_format(e: ft.ControlEvent) -> None:
        page.client_storage.set("auto_detect_input_format", e.control.value)

    if not (page.client_storage.contains_key("reset_tasks_on_input_change")):
        page.client_storage.set("reset_tasks_on_input_change", True)

    def change_reset_tasks_on_input_change(e: ft.ControlEvent) -> None:
        page.client_storage.set("reset_tasks_on_input_change", e.control.value)

    if not (page.client_storage.contains_key("max_track_count")):
        page.client_storage.set("max_track_count", 1)

    def change_max_track_count(e: ft.ControlEvent) -> None:
        page.client_storage.set("max_track_count", e.control.value)

    def on_upload_progress(e: ft.FilePickerUploadEvent) -> None:
        if e.progress == 1.0:
            auto_detect_input_format = page.client_storage.get("auto_detect_input_format")
            last_input_format = page.client_storage.get("last_input_format")
            file_path = settings.save_folder / e.file_name
            suffix = file_path.suffix.lower().removeprefix(".")
            if (
                suffix != last_input_format
                and auto_detect_input_format
                and suffix in plugin_manager.plugin_registry
            ):
                set_last_input_format(suffix)
            task_list_view.current.controls.append(
                ft.ListTile(
                    leading=ft.Stack(
                        [
                            ft.Icon(
                                ft.Icons.ACCESS_TIME_FILLED_OUTLINED,
                                color=ft.colors.GREY_400,
                            ),
                            ft.ProgressRing(visible=False),
                        ]
                    ),
                    title=ft.Text(e.file_name),
                    subtitle=ft.Text(file_path.stem),
                    trailing=ft.PopupMenuButton(
                        items=[
                            ft.PopupMenuItem(
                                icon=ft.icons.REMOVE_RED_EYE_OUTLINED,
                                text=_("View Log"),
                                on_click=show_task_log,
                            ),
                            ft.PopupMenuItem(
                                icon=ft.icons.EDIT,
                                text=_("Rename"),
                                on_click=open_rename_dialog,
                            ),
                            ft.PopupMenuItem(
                                icon=ft.icons.DELETE_OUTLINE,
                                text=_("Remove"),
                                on_click=remove_task,
                            ),
                        ],
                        tooltip=_("Actions"),
                    ),
                    data={"path": file_path, "log_text": ""},
                )
            )
            task_list_view.current.update()

    file_picker = ft.FilePicker(on_result=on_files_selected, on_upload=on_upload_progress)
    permission_handler = fph.PermissionHandler()
    page.overlay.extend([file_picker, permission_handler])

    def check_permission(e: ft.ControlEvent) -> None:
        result: fph.PermissionStatus | None = permission_handler.check_permission(e.control.data)
        banner_ref = ft.Ref[ft.Banner]()
        dismiss_btn = ft.TextButton(text=_("OK"), on_click=lambda _: page.close(banner_ref.current))
        if result != fph.PermissionStatus.GRANTED:
            result = permission_handler.request_permission(e.control.data)
            if result == fph.PermissionStatus.GRANTED:
                banner = ft.Banner(
                    ft.Text(_("Permission granted, you can now select files from your device.")),
                    ref=banner_ref,
                    leading=ft.Icon(ft.Icons.CHECK_CIRCLE_OUTLINED),
                    actions=[dismiss_btn],
                )
            else:
                banner = ft.Banner(
                    ft.Text(
                        _(
                            "Permission denied, please grant the permission to manage external storage."
                        )
                    ),
                    ref=banner_ref,
                    leading=ft.Icon(ft.Icons.WARNING_OUTLINED),
                    actions=[dismiss_btn],
                )
        else:
            banner = ft.Banner(
                ft.Text(
                    _("Permission already granted, you can now select files from your device.")
                ),
                ref=banner_ref,
                leading=ft.Icon(ft.Icons.CHECK_CIRCLE_OUTLINED),
                actions=[dismiss_btn],
            )
        page.open(banner)

    def open_rename_dialog(e: ft.ControlEvent) -> None:
        list_tile = e.control.parent.parent

        def close_rename_dialog() -> None:
            list_tile.subtitle.value = rename_dialog.content.value
            list_tile.subtitle.update()
            page.close(rename_dialog)

        rename_dialog: ft.AlertDialog = ft.AlertDialog(
            title=ft.Text(_("Rename")),
            content=ft.TextField(list_tile.subtitle.value),
            actions=[
                ft.TextButton(_("OK"), on_click=lambda _: close_rename_dialog()),
            ],
            actions_alignment=ft.MainAxisAlignment.END,
        )
        page.open(rename_dialog)

    def remove_task(e: ft.ControlEvent) -> None:
        task_list_view.current.controls.remove(e.control.parent.parent)
        task_list_view.current.update()

    def convert_one(list_tile: ft.ListTile, *sub_tasks: list[ft.ListTile]) -> None:
        conversion_mode = conversion_mode_select.current.value
        if (
            (input_format := page.client_storage.get("last_input_format")) is None
            or (output_format := page.client_storage.get("last_output_format")) is None
            or (max_track_count := page.client_storage.get("max_track_count")) is None
            or (
                save_folder_str := (
                    settings.save_folder if page.web else page.client_storage.get("save_folder")
                )
            )
            is None
            or list_tile.leading is None
            or list_tile.subtitle is None
            or list_tile.data is None
            or list_tile.trailing is None
        ):
            return
        save_folder = pathlib.Path(save_folder_str)
        list_tile.data["log_text"] = ""
        list_tile.leading.controls[0].visible = False
        list_tile.leading.controls[1].visible = True
        list_tile.trailing.items[-1].disabled = True
        list_tile.update()
        try:
            with CatchWarnings() as w:
                output_path = temp_path / list_tile.subtitle.value
                if conversion_mode != "split":
                    output_path = output_path.with_suffix(
                        f".{output_format}",
                    )
                input_plugin = plugin_manager.plugin_registry[input_format]
                output_plugin = plugin_manager.plugin_registry[output_format]
                if (
                    input_plugin.plugin_object is not None
                    and (
                        input_option_class := get_type_hints(input_plugin.plugin_object.load).get(
                            "options",
                        )
                    )
                    is not None
                    and output_plugin.plugin_object is not None
                    and (
                        output_option_class := get_type_hints(
                            output_plugin.plugin_object.dump,
                        ).get("options")
                    )
                    is not None
                ):
                    input_option = input_option_class.model_validate(
                        {
                            control.data: control.value
                            for control in input_options.current.controls
                            if control.data is not None and hasattr(control, "value")
                        }
                    )
                    if conversion_mode == "merge":
                        child_projects = [
                            input_plugin.plugin_object.load(
                                sub_task.data["path"],
                                input_option,
                            )
                            for sub_task in more_itertools.value_chain(list_tile, sub_tasks)
                            if sub_task.data is not None
                        ]
                        project = Project.merge_projects(child_projects)
                    else:
                        project = input_plugin.plugin_object.load(
                            list_tile.data["path"],
                            input_option,
                        )
                    for middleware_id, middleware_ref in middleware_options.items():
                        if (
                            middleware_ref.current.header is not None
                            and middleware_ref.current.header.leading.value
                        ):
                            middleware = middleware_manager.plugin_registry[middleware_id]
                            if (
                                middleware_ref.current.content is not None
                                and middleware.plugin_object is not None
                                and hasattr(middleware.plugin_object, "process")
                                and (
                                    middleware_option_cls := get_type_hints(
                                        middleware.plugin_object.process
                                    ).get(
                                        "options",
                                    )
                                )
                            ):
                                project = middleware.plugin_object.process(
                                    project,
                                    middleware_option_cls.model_validate(
                                        {
                                            control.data: control.value
                                            for control in middleware_ref.current.content.controls[
                                                -1
                                            ].controls
                                            if control.data is not None
                                            and hasattr(control, "value")
                                        }
                                    ),
                                )
                    output_option = output_option_class.model_validate(
                        {
                            control.data: control.value
                            for control in output_options.current.controls
                            if control.data is not None and hasattr(control, "value")
                        }
                    )
                    if conversion_mode == "split":
                        output_path.mkdir(parents=True, exist_ok=True)
                        for i, child_project in enumerate(project.split_tracks(max_track_count)):
                            output_plugin.plugin_object.dump(
                                output_path
                                / f"{list_tile.subtitle.value}_{i + 1:0=2d}.{output_format}",
                                child_project,
                                output_option,
                            )
                    else:
                        output_plugin.plugin_object.dump(
                            output_path,
                            project,
                            output_option,
                        )
            if w.output:
                list_tile.data["log_text"] = w.output
            buffer = io.BytesIO()
            if output_path.is_file():
                buffer.write(output_path.read_bytes())
            else:
                with ZipFile(buffer, "w") as zip_file:
                    for child_file in output_path.iterdir():
                        if not child_file.is_file():
                            continue
                        zip_file.writestr(
                            child_file.name,
                            child_file.read_bytes(),
                        )
            if conversion_mode != "split":
                save_path = save_folder / output_path.name
            else:
                save_path = save_folder / f"{list_tile.subtitle.value}.zip"
            save_path.write_bytes(buffer.getvalue())
            if page.web:
                page.launch_url(f"/download/{save_path.name}")
            if w.output:
                list_tile.leading.controls[0].name = ft.Icons.WARNING_OUTLINED
                list_tile.leading.controls[0].color = ft.Colors.YELLOW_400
            else:
                list_tile.leading.controls[0].name = ft.Icons.CHECK_CIRCLE_OUTLINED
                list_tile.leading.controls[0].color = ft.Colors.GREEN_400
        except Exception:
            list_tile.leading.controls[0].name = ft.Icons.ERROR_OUTLINED
            list_tile.leading.controls[0].color = ft.Colors.RED_400
            list_tile.data["log_text"] = traceback.format_exc()
        list_tile.leading.controls[0].visible = True
        list_tile.leading.controls[1].visible = False
        list_tile.trailing.items[-1].disabled = False
        list_tile.update()

    def on_route_change(event: ft.RouteChangeEvent) -> None:
        page.views.clear()

        def click_navigation_bar(event: ft.ControlEvent | None) -> None:
            for index, p in enumerate(pages.controls):
                p.visible = index == bottom_nav_bar.selected_index
            page.update()

        def convert_all(e: ft.ControlEvent) -> None:
            bottom_nav_bar.selected_index = 1
            click_navigation_bar(None)
            if conversion_mode_select.current.value != "merge":
                for list_tile in task_list_view.current.controls:
                    page.run_thread(convert_one, list_tile)
            elif len(task_list_view.current.controls) > 0:
                page.run_thread(
                    convert_one,
                    task_list_view.current.controls[0],
                    *task_list_view.current.controls[1:],
                )

        bottom_nav_bar = ft.NavigationBar(
            on_change=click_navigation_bar,
            destinations=[
                ft.NavigationBarDestination(
                    icon=ft.Icons.SWAP_HORIZ_OUTLINED, label=_("Select File Formats")
                ),
                ft.NavigationBarDestination(
                    icon=ft.Icons.TASK_ALT_OUTLINED, label=_("Conversion mode & Task list")
                ),
                ft.NavigationBarDestination(
                    icon=ft.Icons.SETTINGS_OUTLINED, label=_("Advanced Options")
                ),
                ft.NavigationBarDestination(
                    icon=ft.Icons.SETTINGS_APPLICATIONS_OUTLINED, label=_("Basic Settings")
                ),
                ft.NavigationBarDestination(icon=ft.Icons.HELP_OUTLINED, label=_("About")),
            ],
        )
        switch_theme_btn = ft.PopupMenuButton(
            icon=ft.Icons.PALETTE_OUTLINED,
            tooltip=_("Switch Theme"),
            items=[
                ft.PopupMenuItem(
                    text=_("System"),
                    icon=ft.Icons.BRIGHTNESS_AUTO_OUTLINED,
                    on_click=lambda _: change_theme("System"),
                ),
                ft.PopupMenuItem(
                    text=_("Light"),
                    icon=ft.Icons.LIGHT_MODE_OUTLINED,
                    on_click=lambda _: change_theme("Light"),
                ),
                ft.PopupMenuItem(
                    text=_("Dark"),
                    icon=ft.Icons.DARK_MODE_OUTLINED,
                    on_click=lambda _: change_theme("Dark"),
                ),
            ],
        )
        switch_language_btn = ft.PopupMenuButton(
            icon=ft.Icons.TRANSLATE_OUTLINED,
            tooltip=_("Switch Language"),
            items=[
                ft.PopupMenuItem(text="简体中文", on_click=lambda _: change_language("zh_CN")),
                ft.PopupMenuItem(text="English", on_click=lambda _: change_language("en_US")),
                ft.PopupMenuItem(text="Deutsch", on_click=lambda _: change_language("de_DE")),
            ],
        )
        window_buttons = []
        if page.platform not in [ft.PagePlatform.IOS, ft.PagePlatform.ANDROID] and not page.web:
            maximize_button = ft.Ref[ft.IconButton]()

            def on_minimize_click(e: ft.ControlEvent) -> None:
                page.window.minimized = True
                page.update()

            def on_maximize_click(e: ft.ControlEvent) -> None:
                page.window.maximized = not page.window.maximized
                page.update()
                if page.window.maximized:
                    maximize_button.current.icon = ft.Icons.CLOSE_FULLSCREEN_OUTLINED
                    maximize_button.current.tooltip = _("Restore")
                else:
                    maximize_button.current.icon = ft.Icons.OPEN_IN_FULL_OUTLINED
                    maximize_button.current.tooltip = _("Maximize")
                maximize_button.current.update()

            window_buttons.extend(
                [
                    ft.IconButton(
                        icon=ft.Icons.MINIMIZE_OUTLINED,
                        tooltip=_("Minimize"),
                        on_click=on_minimize_click,
                    ),
                    ft.IconButton(
                        ref=maximize_button,
                        icon=ft.Icons.OPEN_IN_FULL_OUTLINED,
                        tooltip=_("Maximize"),
                        on_click=on_maximize_click,
                    ),
                    ft.IconButton(
                        icon=ft.Icons.CLOSE_OUTLINED,
                        hover_color=ft.Colors.RED_400,
                        tooltip=_("Close"),
                        on_click=lambda _: page.window.close(),
                    ),
                ]
            )

        def swap_formats(e: ft.ControlEvent) -> None:
            last_output_format, last_input_format = (
                output_select.current.value,
                input_select.current.value,
            )
            set_last_output_format(last_input_format)
            set_last_input_format(last_output_format)
            page.update()

        def show_plugin_info(control: ft.Ref[ft.Dropdown]) -> None:
            if control.current.value:
                plugin_obj = plugin_manager.plugin_registry[control.current.value]
                page.views.append(
                    ft.View(
                        "/plugin_info",
                        appbar=ft.AppBar(
                            title=ft.Text(plugin_obj.name),
                            center_title=True,
                            bgcolor=ft.Colors.SURFACE,
                            leading=ft.IconButton(
                                ft.Icons.ARROW_BACK_OUTLINED,
                                tooltip=_("Back"),
                                on_click=lambda _: view_pop(None),
                            ),
                        ),
                        controls=[
                            ft.ResponsiveRow(
                                [
                                    ft.Image(
                                        src_base64=plugin_obj.icon_base64,
                                        fit=ft.ImageFit.FILL,
                                        col=3,
                                    ),
                                    ft.ResponsiveRow(
                                        [
                                            ft.Icon(ft.Icons.BOOKMARK_OUTLINE_OUTLINED, col=1),
                                            ft.Text(
                                                str(plugin_obj.version), tooltip=_("Version"), col=3
                                            ),
                                            ft.Icon(ft.Icons.PERSON_OUTLINE_OUTLINED, col=1),
                                            ft.Row(
                                                [
                                                    ft.Text(
                                                        spans=[
                                                            ft.TextSpan(
                                                                _(plugin_obj.author),
                                                                ft.TextStyle(
                                                                    decoration=ft.TextDecoration.UNDERLINE
                                                                ),
                                                                url=plugin_obj.website,
                                                            ),
                                                        ],
                                                        tooltip=plugin_obj.website,
                                                    ),
                                                    ft.Icon(ft.Icons.OPEN_IN_NEW_OUTLINED),
                                                ],
                                                col=7,
                                            ),
                                            ft.Icon(ft.Icons.INSERT_DRIVE_FILE_OUTLINED, col=1),
                                            ft.Text(
                                                f"{_(plugin_obj.file_format)} (*.{plugin_obj.suffix})",
                                                col=11,
                                            ),
                                        ],
                                        col=9,
                                    ),
                                    ft.Divider(),
                                    ft.Text(
                                        spans=[
                                            ft.TextSpan(
                                                _("Introduction"),
                                                ft.TextStyle(weight=ft.FontWeight.BOLD),
                                            ),
                                        ],
                                        col=12,
                                    ),
                                    ft.Text(_(plugin_obj.description)),
                                ],
                            )
                        ],
                    )
                )
                page.update()

        pages = ft.Stack(
            [
                ft.Column(
                    [
                        ft.ResponsiveRow(
                            [
                                ft.Dropdown(
                                    ref=input_select,
                                    value=page.client_storage.get("last_input_format"),
                                    label=_("Import format"),
                                    text_size=14,
                                    options=[
                                        ft.dropdown.Option(
                                            plugin_id,
                                            f"{_(plugin_obj.file_format)} (*.{plugin_obj.suffix})",
                                        )
                                        for plugin_id, plugin_obj in plugin_manager.plugin_registry.items()
                                    ],
                                    col=10,
                                    dense=True,
                                    on_change=lambda e: set_last_input_format(e.control.value),
                                ),
                                ft.IconButton(
                                    icon=ft.Icons.INFO_OUTLINE,
                                    tooltip=_("View Detail Information"),
                                    col=2,
                                    on_click=lambda _: show_plugin_info(input_select),
                                ),
                                ft.Container(col=10),
                                ft.IconButton(
                                    icon=ft.Icons.SWAP_VERT_CIRCLE_OUTLINED,
                                    tooltip=_("Swap Input and Output"),
                                    col=2,
                                    on_click=swap_formats,
                                ),
                                ft.Dropdown(
                                    ref=output_select,
                                    value=page.client_storage.get("last_output_format"),
                                    label=_("Export format"),
                                    text_size=14,
                                    options=[
                                        ft.dropdown.Option(
                                            plugin_id,
                                            f"{_(plugin_obj.file_format)} (*.{plugin_obj.suffix})",
                                        )
                                        for plugin_id, plugin_obj in plugin_manager.plugin_registry.items()
                                    ],
                                    col=10,
                                    dense=True,
                                    on_change=lambda e: set_last_output_format(e.control.value),
                                ),
                                ft.IconButton(
                                    icon=ft.Icons.INFO_OUTLINE,
                                    tooltip=_("View Detail Information"),
                                    col=2,
                                    on_click=lambda _: show_plugin_info(output_select),
                                ),
                            ],
                            vertical_alignment=ft.CrossAxisAlignment.CENTER,
                        ),
                    ],
                    visible=True,
                ),
                ft.Column(
                    [
                        ft.Dropdown(
                            ref=conversion_mode_select,
                            value="direct",
                            label=_("Conversion mode"),
                            options=[
                                ft.dropdown.Option("direct", _("Direct")),
                                ft.dropdown.Option("split", _("Split")),
                                ft.dropdown.Option("merge", _("Merge")),
                            ],
                        ),
                        ft.ListView(
                            ref=task_list_view,
                            expand=1,
                            spacing=10,
                            auto_scroll=True,
                            padding=ft.padding.symmetric(vertical=10),
                        ),
                    ],
                    visible=False,
                ),
                ft.Column(
                    [
                        ft.ExpansionPanelList(
                            elevation=8,
                            controls=[
                                ft.ExpansionPanel(
                                    header=ft.ListTile(
                                        leading=ft.Icon(ft.Icons.INPUT),
                                        title=ft.Text(_("Input Options")),
                                    ),
                                    content=ft.Column(
                                        [
                                            ft.Container(height=2),
                                            ft.ResponsiveRow(
                                                controls=build_input_options(
                                                    page.client_storage.get("last_input_format")
                                                ),
                                                ref=input_options,
                                                vertical_alignment=ft.CrossAxisAlignment.CENTER,
                                            ),
                                        ]
                                    ),
                                ),
                                *(
                                    ft.ExpansionPanel(
                                        ref=middleware_ref,
                                        header=ft.ListTile(
                                            leading=ft.Switch(value=False),
                                            title=ft.Text(
                                                _(
                                                    middleware_manager.plugin_registry[
                                                        middleware_id
                                                    ].name
                                                )
                                            ),
                                        ),
                                        content=ft.Column(
                                            [
                                                ft.Container(height=2),
                                                ft.ResponsiveRow(
                                                    controls=build_middleware_options(
                                                        middleware_id
                                                    ),
                                                    vertical_alignment=ft.CrossAxisAlignment.CENTER,
                                                ),
                                            ]
                                        ),
                                    )
                                    for middleware_id, middleware_ref in middleware_options.items()
                                ),
                                ft.ExpansionPanel(
                                    header=ft.ListTile(
                                        leading=ft.Icon(ft.Icons.OUTPUT),
                                        title=ft.Text(_("Output Options")),
                                    ),
                                    content=ft.Column(
                                        [
                                            ft.Container(height=2),
                                            ft.ResponsiveRow(
                                                controls=build_output_options(
                                                    page.client_storage.get("last_output_format")
                                                ),
                                                ref=output_options,
                                                vertical_alignment=ft.CrossAxisAlignment.CENTER,
                                            ),
                                        ],
                                    ),
                                ),
                            ],
                        )
                    ],
                    scroll=ft.ScrollMode.ALWAYS,
                    height=400,
                    visible=False,
                ),
                ft.ResponsiveRow(
                    [
                        ft.Switch(
                            _("Auto detect import format"),
                            value=page.client_storage.get("auto_detect_input_format"),
                            col=12,
                            on_change=change_auto_detect_input_format,
                        ),
                        ft.Switch(
                            _("Reset list when import format changed"),
                            value=page.client_storage.get("reset_tasks_on_input_change"),
                            col=12,
                            on_change=change_reset_tasks_on_input_change,
                        ),
                        ft.Text(_("Max track count"), col=4),
                        ft.Slider(
                            value=page.client_storage.get("max_track_count"),
                            min=1,
                            max=100,
                            divisions=100,
                            label="{value}",
                            col=8,
                            on_change=change_max_track_count,
                        ),
                        ft.TextField(
                            ref=save_folder_text_field,
                            label=_("Output Folder"),
                            value=page.client_storage.get("save_folder"),
                            col=10,
                        ),
                        ft.IconButton(
                            ft.Icons.FOLDER_OPEN_OUTLINED,
                            tooltip=_("Change Output Directory"),
                            col=2,
                            on_click=lambda e: file_picker.get_directory_path(
                                _("Change Output Directory")
                            ),
                        ),
                        ft.ElevatedButton(
                            _("Request permission to access files"),
                            data=fph.PermissionType.MANAGE_EXTERNAL_STORAGE,
                            on_click=check_permission,
                            col=12,
                        ),
                    ],
                    vertical_alignment=ft.CrossAxisAlignment.CENTER,
                    visible=False,
                ),
                ft.ResponsiveRow(
                    [
                        ft.Text(
                            spans=[
                                ft.TextSpan(
                                    "LibreSVIP",
                                    ft.TextStyle(size=24, weight=ft.FontWeight.BOLD),
                                ),
                            ],
                            text_align=ft.TextAlign.CENTER,
                            col=12,
                        ),
                        ft.Text(
                            _("Version: ") + libresvip.__version__,
                            text_align=ft.TextAlign.CENTER,
                            col=12,
                        ),
                        ft.Text(_("Author: SoulMelody"), text_align=ft.TextAlign.CENTER, col=12),
                        ft.TextButton(
                            _("Author's Profile"),
                            icon=ft.Icons.LIVE_TV_OUTLINED,
                            url="https://space.bilibili.com/175862486",
                            col=6,
                        ),
                        ft.TextButton(
                            _("Repo URL"),
                            icon=ft.Icons.LOGO_DEV_OUTLINED,
                            url="https://github.com/SoulMelody/LibreSVIP",
                            col=6,
                        ),
                        ft.Card(
                            ft.Column(
                                [
                                    ft.Text(
                                        _(
                                            "LibreSVIP is an open-sourced, liberal and extensionable framework that can convert your singing synthesis projects between different file formats.",
                                        )
                                    ),
                                    ft.Text(
                                        _(
                                            "All people should have the right and freedom to choose. That's why we're committed to giving you a second chance to keep your creations free from the constraints of platforms and coterie.",
                                        )
                                    ),
                                ]
                            ),
                            col=12,
                        ),
                    ],
                    visible=False,
                ),
            ]
        )

        page.views.append(
            ft.View(
                "/",
                appbar=ft.AppBar(
                    title=ft.WindowDragArea(
                        ft.Row([ft.Text("LibreSVIP")], expand=True), expand=True
                    ),
                    bgcolor=ft.Colors.SURFACE,
                    actions=[
                        ft.IconButton(
                            ft.Icons.ADD_OUTLINED,
                            tooltip=_("Continue Adding files"),
                            on_click=lambda e: file_picker.pick_files(
                                _("Select files to convert"), allow_multiple=True
                            ),
                        ),
                        switch_theme_btn,
                        switch_language_btn,
                        *window_buttons,
                    ],
                ),
                navigation_bar=bottom_nav_bar,
                floating_action_button=ft.FloatingActionButton(
                    icon=ft.Icons.PLAY_ARROW_OUTLINED,
                    tooltip=_("Start Conversion"),
                    on_click=convert_all,
                ),
                controls=[pages],
            )
        )
        page.update()

    def view_pop(view: ft.View | None) -> None:
        page.views.pop()
        top_view = page.views[-1]
        page.go(top_view.route or "/")

    def on_view_pop(event: ft.ViewPopEvent) -> None:
        view_pop(event.view)

    def on_keyboard_event(event: ft.KeyboardEvent) -> None:
        if event.key == "Escape" and len(page.views) > 1:
            view_pop(page.views[-1])

    page.on_view_pop = on_view_pop
    page.on_keyboard_event = on_keyboard_event
    page.on_route_change = on_route_change
    page.go("/")



from typing import Final

DEFAULT_CHINESE_PHONEME: Final[str] = "l a"
DEFAULT_JAPANESE_PHONEME: Final[str] = "4 a"
PITCH_BEND_NAME: Final[str] = "pitchBend"
PITCH_BEND_SENSITIVITY_NAME: Final[str] = "pitchBendSens"
BPM_RATE: Final[float] = 100.0



from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note
from libresvip.utils.music_math import midi2note

from . import lyric_util, pinyin_util
from .models.ds_note import (
    AspirationDsNote,
    AspirationDsPhoneme,
    DsNote,
    DsPhoneme,
    DsPhonemeItem,
    RestDsNote,
    RestDsPhoneme,
)


def encode_notes(
    os_notes: list[Note],
    synchronizer: TimeSynchronizer,
    trailing_space: float = 0.05,
) -> list[DsNote]:
    ds_notes: list[DsNote] = []
    prev_end_in_ticks = 0
    prev_actual_end_in_secs = 0.0
    prev_phoneme = DsPhoneme()
    init_pinyin_utils(os_notes)
    min_asp_len = 0.1
    max_asp_len = 0.4
    for index, note in enumerate(os_notes):
        # Calculate Positions
        prev_end_in_secs = synchronizer.get_actual_secs_from_ticks(prev_end_in_ticks)
        cur_start_in_ticks = note.start_pos
        cur_end_in_ticks = cur_start_in_ticks + note.length
        cur_start_in_secs = synchronizer.get_actual_secs_from_ticks(cur_start_in_ticks)
        cur_end_in_secs = synchronizer.get_actual_secs_from_ticks(cur_end_in_ticks)
        cur_actual_start_in_secs = cur_start_in_secs
        cur_actual_end_in_secs = cur_end_in_secs
        if note.edited_phones is not None and note.edited_phones.head_length_in_secs >= 0:
            cur_actual_start_in_secs -= note.edited_phones.head_length_in_secs
        if (
            index < len(os_notes) - 1
            and os_notes[index + 1].edited_phones is not None
            and os_notes[index + 1].edited_phones.head_length_in_secs >= 0  # type: ignore[union-attr]
        ):
            next_note = os_notes[index + 1]
            next_start_in_ticks = next_note.start_pos
            next_start_in_secs = synchronizer.get_actual_secs_from_ticks(next_start_in_ticks)
            next_head = os_notes[index + 1].edited_phones.head_length_in_secs  # type: ignore[union-attr]
            next_actual_start_in_secs = next_start_in_secs - next_head
            if cur_end_in_secs > next_actual_start_in_secs:
                cur_actual_end_in_secs -= cur_end_in_secs - next_actual_start_in_secs

        # Fill Note Gap
        gap = cur_actual_start_in_secs - prev_actual_end_in_secs  # 音符间隙
        if gap > 0:  # 有间隙
            if gap < min_asp_len:  # 间隙很小, 休止
                rest_phoneme = RestDsPhoneme(
                    _duration=round(cur_actual_start_in_secs - prev_actual_end_in_secs, 6)
                )
                rest_note = RestDsNote(
                    round(cur_start_in_secs - prev_end_in_secs, 6),
                    rest_phoneme,
                )
                ds_notes.append(rest_note)
                prev_phoneme = rest_phoneme
            elif gap < max_asp_len:  # 间隙适中, 换气
                asp_phoneme = AspirationDsPhoneme(
                    _duration=round(cur_actual_start_in_secs - prev_actual_end_in_secs, 6)
                )
                aps_note = AspirationDsNote(
                    round(cur_start_in_secs - prev_end_in_secs, 6), asp_phoneme
                )
                ds_notes.append(aps_note)
                prev_phoneme = asp_phoneme
            else:  # 间隙很大, 换气
                rest_phoneme = RestDsPhoneme(
                    _duration=round(
                        cur_actual_start_in_secs - prev_actual_end_in_secs - max_asp_len,
                        6,
                    )
                )
                rest_note = RestDsNote(
                    round(cur_start_in_secs - prev_end_in_secs - max_asp_len, 6),
                    rest_phoneme,
                )
                ds_notes.append(rest_note)
                asp_phoneme = AspirationDsPhoneme(_duration=round(max_asp_len, 6))
                aps_note = AspirationDsNote(round(max_asp_len, 6), asp_phoneme)
                ds_notes.append(aps_note)
                prev_phoneme = asp_phoneme

        # Convert OpenSvip Notes To DsNotes
        ds_phoneme = DsPhoneme()
        if "-" in note.lyric:  # 转音
            ds_phoneme.vowel = DsPhonemeItem(
                phoneme=prev_phoneme.vowel.phoneme,
                duration=round(cur_actual_end_in_secs - cur_actual_start_in_secs, 6),
                note_name=midi2note(note.key_number),
            )
        else:
            pinyin = note.pronunciation or pinyin_util.get_note_pinyin(note.lyric, index)
            consonant, vowel = pinyin_util.split(pinyin)
            if consonant:  # 不是纯元音
                consonant_note_name = (
                    midi2note(note.key_number)
                    if prev_phoneme.vowel.phoneme in ["SP", "AP"]
                    else prev_phoneme.vowel.note_name
                )
                ds_phoneme.consonant = DsPhonemeItem(
                    phoneme=consonant,
                    duration=None
                    if cur_actual_start_in_secs == cur_start_in_secs
                    else round(cur_start_in_secs - cur_actual_start_in_secs, 6),
                    note_name=consonant_note_name,
                )
                ds_phoneme.vowel = DsPhonemeItem(
                    phoneme=vowel,
                    duration=round(cur_actual_end_in_secs - cur_start_in_secs, 6),
                    note_name=midi2note(note.key_number),
                )
            else:  # 纯元音
                ds_phoneme.vowel = DsPhonemeItem(
                    phoneme=vowel,
                    duration=round(cur_actual_end_in_secs - cur_actual_start_in_secs, 6),
                    note_name=midi2note(note.key_number),
                )
        ds_note = DsNote(
            lyric=lyric_util.get_symbol_removed_lyric(note.lyric),
            ds_phoneme=ds_phoneme,
            note_name=midi2note(note.key_number),
            duration=round(cur_end_in_secs - cur_start_in_secs, 6),
        )
        ds_notes.append(ds_note)

        prev_end_in_ticks = cur_end_in_ticks
        prev_actual_end_in_secs = cur_actual_end_in_secs
        prev_phoneme = ds_phoneme
    insert_end_rest_note(ds_notes, trailing_space)
    return ds_notes


def insert_end_rest_note(ds_notes: list[DsNote], trailing_space: float = 0.05) -> None:
    end_rest_phoneme = RestDsPhoneme(_duration=trailing_space)
    end_rest_note = RestDsNote(duration=trailing_space, ds_phoneme=end_rest_phoneme)
    ds_notes.append(end_rest_note)


def init_pinyin_utils(os_notes: list[Note]) -> None:
    lyric_list = [note.lyric for note in os_notes]
    pinyin_util.clear_all_pinyin()
    pinyin_util.add_pinyin_from_lyrics(lyric_list)



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.translation import gettext_lazy as _

from .binsvip_generator import BinarySvipGenerator
from .binsvip_parser import BinarySvipParser
from .msnrbf.svip_reader import SvipReader
from .msnrbf.svip_writer import SvipWriter
from .options import BinarySvipVersion, InputOptions, OutputOptions


class SvipConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        with SvipReader() as reader:
            version, xs_project = reader.read(path)
            return BinarySvipParser(options).parse_project(version, xs_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ver_enum = options.version
        if ver_enum == BinarySvipVersion.SVIP7_0_0:
            project.version = "SVIP7.0.0"
        elif ver_enum == BinarySvipVersion.AUTO:
            if project.version == "SVIP0.0.0":
                project.version = "SVIP6.0.0"
        elif ver_enum == BinarySvipVersion.SVIP6_0_0:
            project.version = "SVIP6.0.0"
        elif ver_enum == BinarySvipVersion.COMPAT:
            project.version = "SVIP0.0.0"
        else:
            raise ValueError(_("Unexpected enum value"))
        with SvipWriter() as registry:
            version, xs_project = BinarySvipGenerator(options=options).generate_project(project)
            registry.write(path, version, xs_project)



import dataclasses

from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point

from .model import (
    AISAudioTrack,
    AISNote,
    AISProjectBody,
    AISProjectHead,
    AISSingVoiceTrack,
    AISTempo,
    AISTimeSignature,
    AISTrack,
)
from .options import InputOptions


@dataclasses.dataclass
class AiSingersParser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, ais_head: AISProjectHead, ais_body: AISProjectBody) -> Project:
        time_signatures = self.parse_time_signatures(ais_head.signature)
        self.first_bar_length = round(time_signatures[0].bar_length())
        tempos = self.parse_tempos(ais_head.tempo)
        track_list = self.parse_tracks(ais_body.tracks)
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=track_list,
        )

    def parse_time_signatures(
        self, ais_time_signatures: list[AISTimeSignature]
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                numerator=ais_time_signature.beat_zi,
                denominator=ais_time_signature.beat_mu,
                bar_index=ais_time_signature.start_bar,
            )
            for ais_time_signature in ais_time_signatures
        ]

    def parse_tempos(self, ais_tempos: list[AISTempo]) -> list[SongTempo]:
        return [
            SongTempo(
                bpm=ais_tempo.tempo_float,
                position=ais_tempo.start_128 * 15,
            )
            for ais_tempo in ais_tempos
        ]

    def parse_tracks(self, ais_tracks: list[AISTrack]) -> list[Track]:
        track_list = []
        for ais_track in ais_tracks:
            if isinstance(ais_track, AISSingVoiceTrack):
                note_list = []
                pitch_points = [Point.start_point()]
                for item in ais_track.items:
                    tick_prefix = item.start * 15
                    item_notes, pitch_points = self.parse_notes(item.notes, tick_prefix)
                    note_list.extend(item_notes)
                    pitch_points.extend(pitch_points)
                pitch_points.append(Point.end_point())
                singing_track = SingingTrack(
                    title=ais_track.name,
                    ai_singer_name=ais_track.singer_namecn,
                    mute=ais_track.mute,
                    solo=ais_track.solo,
                    note_list=note_list,
                )
                if len(pitch_points) > 2:
                    singing_track.edited_params.pitch = ParamCurve(points=Points(root=pitch_points))
                track_list.append(singing_track)
            elif self.options.import_instrumental_track and isinstance(ais_track, AISAudioTrack):
                track_list.extend(
                    InstrumentalTrack(
                        title=f"{ais_track.name} ({i})",
                        mute=ais_track.mute,
                        solo=ais_track.solo,
                        audio_file_path=item.path_audio,
                        offset=item.start * 15,
                    )
                    for i, item in enumerate(ais_track.items)
                )
        return track_list

    def parse_notes(
        self, ais_notes: list[AISNote], tick_prefix: int
    ) -> tuple[list[Note], list[Point]]:
        note_list = []
        pitch_points = []
        for ais_note in ais_notes:
            note = Note(
                start_pos=ais_note.start * 15 + tick_prefix,
                length=ais_note.length * 15,
                key_number=ais_note.midi_no + 12,
                lyric=ais_note.lyric,
                pronunciation=ais_note.pinyin,
            )
            if self.options.import_pitch:
                tick_step = note.length / len(ais_note.pit)
                pitch_points.append(Point(x=note.start_pos + self.first_bar_length, y=-100))
                for i in range(len(ais_note.pit)):
                    tick = note.start_pos + tick_step * i
                    pitch_points.append(
                        Point(
                            x=round(tick) + self.first_bar_length,
                            y=round((note.key_number) * 100 + ais_note.pit[i] * 10),
                        )
                    )
                pitch_points.append(Point(x=note.end_pos + self.first_bar_length, y=-100))
            note_list.append(note)
        return note_list, pitch_points



import functools

from fonticon_mdi7 import MDI7
from PySide6.QtCore import QObject, QUrl, Slot
from PySide6.QtQml import QmlElement

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


@QmlElement
class IconicFontLoader(QObject):
    @Slot(str, result=str)
    def font_path(self, font_family: str) -> str:
        return QUrl.from_local_file(MDI7.__font_file__).to_string()

    @Slot(str, result=str)
    @functools.cache
    def icon(self, icon_name: str) -> str:
        font_family, _, icon_name = icon_name.partition(".")
        icon_name = icon_name.replace("-", "_")
        return getattr(MDI7, icon_name, "").partition(".")[-1]
